import dataclasses
import urllib
from typing import Any, Optional, Union

from ..trace_server import refs_internal

DICT_KEY_EDGE_NAME = refs_internal.DICT_KEY_EDGE_NAME
LIST_INDEX_EDGE_NAME = refs_internal.LIST_INDEX_EDGE_NAME
OBJECT_ATTR_EDGE_NAME = refs_internal.OBJECT_ATTR_EDGE_NAME
TABLE_ROW_ID_EDGE_NAME = refs_internal.TABLE_ROW_ID_EDGE_NAME


@dataclasses.dataclass(frozen=True)
class Ref:
    def uri(self) -> str:
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class TableRef(Ref):
    entity: str
    project: str
    digest: str
    row_digests: Optional[list[str]] = None

    def uri(self) -> str:
        return f"weave:///{self.entity}/{self.project}/table/{self.digest}"


@dataclasses.dataclass(frozen=True)
class RefWithExtra(Ref):
    def with_extra(self, extra: tuple[str, ...]) -> "RefWithExtra":
        params = dataclasses.asdict(self)
        params["extra"] = self.extra + tuple(extra)  # type: ignore
        return self.__class__(**params)

    def with_key(self, key: str) -> "RefWithExtra":
        return self.with_extra((DICT_KEY_EDGE_NAME, key))

    def with_attr(self, attr: str) -> "RefWithExtra":
        return self.with_extra((OBJECT_ATTR_EDGE_NAME, attr))

    def with_index(self, index: int) -> "RefWithExtra":
        return self.with_extra((LIST_INDEX_EDGE_NAME, str(index)))

    def with_item(self, item_digest: str) -> "RefWithExtra":
        return self.with_extra((TABLE_ROW_ID_EDGE_NAME, f"{item_digest}"))


@dataclasses.dataclass(frozen=True)
class ObjectRef(RefWithExtra):
    entity: str
    project: str
    name: str
    digest: str
    extra: tuple[str, ...] = ()

    def __post_init__(self) -> None:
        refs_internal.validate_no_slashes(self.digest, "digest")
        refs_internal.validate_no_colons(self.digest, "digest")
        refs_internal.validate_extra(list(self.extra))
        refs_internal.validate_no_slashes(self.name, "name")
        refs_internal.validate_no_colons(self.name, "name")

    def uri(self) -> str:
        u = f"weave:///{self.entity}/{self.project}/object/{self.name}:{self.digest}"
        if self.extra:
            u += "/" + "/".join(refs_internal.extra_value_quoter(e) for e in self.extra)
        return u

    def get(self) -> Any:
        # Move import here so that it only happens when the function is called.
        # This import is invalid in the trace server and represents a dependency
        # that should be removed.
        from weave.trace.client_context.weave_client import get_weave_client
        from weave.trace.weave_init import init_weave

        gc = get_weave_client()
        if gc is not None:
            return gc.get(self)

        # Special case: If the user is attempting to fetch an object but has not
        # yet initialized the client, we can initialize a client to
        # fetch the object. It is critical to reset the client after fetching the
        # object to avoid any side effects in user code.
        if gc is None:
            init_client = init_weave(
                f"{self.entity}/{self.project}", ensure_project_exists=False
            )
            try:
                res = init_client.client.get(self)
            finally:
                init_client.reset()
            return res

    def is_descended_from(self, potential_ancestor: "ObjectRef") -> bool:
        if self.entity != potential_ancestor.entity:
            return False
        if self.project != potential_ancestor.project:
            return False
        if self.name != potential_ancestor.name:
            return False
        if self.digest != potential_ancestor.digest:
            return False
        if len(self.extra) <= len(potential_ancestor.extra):
            return False
        return all(
            self.extra[i] == potential_ancestor.extra[i]
            for i in range(len(potential_ancestor.extra))
        )


@dataclasses.dataclass(frozen=True)
class OpRef(ObjectRef):
    def uri(self) -> str:
        u = f"weave:///{self.entity}/{self.project}/op/{self.name}:{self.digest}"
        if self.extra:
            u += "/" + "/".join(refs_internal.extra_value_quoter(e) for e in self.extra)
        return u


@dataclasses.dataclass(frozen=True)
class CallRef(RefWithExtra):
    entity: str
    project: str
    id: str
    extra: tuple[str, ...] = ()

    def uri(self) -> str:
        u = f"weave:///{self.entity}/{self.project}/call/{self.id}"
        if self.extra:
            u += "/" + "/".join(refs_internal.extra_value_quoter(e) for e in self.extra)
        return u


AnyRef = Union[ObjectRef, TableRef, CallRef, OpRef]


def parse_uri(uri: str) -> AnyRef:
    if not uri.startswith("weave:///"):
        raise ValueError(f"Invalid URI: {uri}")
    path = uri[len("weave:///") :]
    parts = path.split("/")
    if len(parts) < 3:
        raise ValueError(f"Invalid URI: {uri}")
    entity, project, kind = parts[:3]
    remaining = tuple(parts[3:])
    if kind == "table":
        return TableRef(entity=entity, project=project, digest=remaining[0])
    extra = tuple(urllib.parse.unquote(r) for r in remaining[1:])
    if kind == "call":
        return CallRef(entity=entity, project=project, id=remaining[0], extra=extra)
    elif kind == "object":
        name, version = remaining[0].split(":")
        return ObjectRef(
            entity=entity, project=project, name=name, digest=version, extra=extra
        )
    elif kind == "op":
        name, version = remaining[0].split(":")
        return OpRef(
            entity=entity, project=project, name=name, digest=version, extra=extra
        )
    else:
        raise ValueError(f"Unknown ref kind: {kind}")


def parse_op_uri(uri: str) -> OpRef:
    if not isinstance(parsed := parse_uri(uri), OpRef):
        raise ValueError(f"URI is not for an Op: {uri}")
    return parsed



import csv

import pyarrow as pa
import pyarrow.csv as pa_csv

from weave.legacy.weave import api as weave
from weave.legacy.weave import file_base


def sniff_dialect(path: str) -> type[csv.Dialect]:
    with open(path, newline="") as csvfile:
        dialect = csv.Sniffer().sniff(csvfile.read(100 * 1024))
        return dialect


def dialect_to_pyarrow_options(
    dialect: type[csv.Dialect],
) -> tuple[pa_csv.ReadOptions, pa_csv.ParseOptions]:
    # Convert a csv.Dialect object to the corresponding pyarrow options.
    read_options = pa_csv.ReadOptions()
    parse_options = pa_csv.ParseOptions(
        delimiter=dialect.delimiter,
        quote_char=dialect.quotechar,
        double_quote=dialect.doublequote,
        escape_char=dialect.escapechar,
        newlines_in_values=True,
    )
    return read_options, parse_options


def read_csv_with_dialect(path: str) -> pa.Table:
    dialect = sniff_dialect(path)
    read_options, parse_options = dialect_to_pyarrow_options(dialect)
    table = pa_csv.read_csv(
        path, read_options=read_options, parse_options=parse_options
    )
    return table


def convert_type(val):
    if val is None:
        return val
    try:
        return int(val)
    except ValueError:
        try:
            return float(val)
        except ValueError:
            return val


def load_csv(csvfile):
    dialect = csv.Sniffer().sniff(csvfile.read(10240), delimiters=";,")
    csvfile.seek(0)
    reader = csv.DictReader(csvfile, dialect=dialect)
    rows = []
    for row in reader:
        # DictReader puts items that don't have a header into a list under
        # the None key. This happens for mal-formed csvs. Ignore the None
        # key.
        rows.append({k: convert_type(v) for k, v in row.items() if k is not None})
    return rows


def save_csv(csvfile, csv_data):
    if not csv_data:
        field_names = []
    else:
        field_names = list(csv_data[0].keys())
    writer = csv.DictWriter(csvfile, field_names, delimiter=";")
    writer.writeheader()
    for row in csv_data:
        writer.writerow(row)


def writecsv(self: file_base.File, csv_data: list[dict]):
    with self.open("w") as f:
        save_csv(f, csv_data)


@weave.op(
    name="file-refine_readcsv",
    hidden=True,
    input_type={"self": file_base.FileBaseType(extension=weave.types.literal("csv"))},
)
def refine_readcsv(self) -> weave.types.Type:
    with self.open() as f:
        return weave.types.TypeRegistry.type_of(load_csv(f))


@weave.op(
    # TODO: I had to mark pure=False
    # But that's not true! We need to know if the file we're reading is
    # immutable (inside an artifact) or not (on a filesystem).
    setter=writecsv,
    name="file-readcsv",
    input_type={"self": file_base.FileBaseType(extension=weave.types.literal("csv"))},
    output_type=weave.types.List(weave.types.TypedDict({})),
    refine_output_type=refine_readcsv,
)
def readcsv(self):
    with self.open() as f:
        return load_csv(f)



from weave.legacy.weave import ref_base
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import graph, mappers, node_ref
from weave.legacy.weave import mappers_python_def as mappers_python


class FunctionToPyFunction(mappers.Mapper):
    def apply(self, obj):
        # TODO: This should be a check on input type rather than a value check!
        #   As implemented it might be possible to save in two different formats
        #   for the save input type, which is bad.
        return obj.to_json()


class PyFunctionToFunction(mappers.Mapper):
    def apply(self, obj):
        if isinstance(obj, str):
            ref = ref_base.Ref.from_str(obj)
            return node_ref.ref_to_node(ref)
        # Obj is graph.Node
        return graph.Node.node_from_json(obj)


mappers_python.add_mapper(types.Function, FunctionToPyFunction, PyFunctionToFunction)



import atexit
import logging
import time
from queue import Queue
from threading import Event, Lock, Thread
from typing import Callable, Generic, List, TypeVar

from weave.trace.context import get_raise_on_captured_errors
from weave.trace_server import requests

T = TypeVar("T")
logger = logging.getLogger(__name__)


class AsyncBatchProcessor(Generic[T]):
    """A class that asynchronously processes batches of items using a provided processor function."""

    def __init__(
        self,
        processor_fn: Callable[[List[T]], None],
        max_batch_size: int = 100,
        min_batch_interval: float = 1.0,
    ) -> None:
        """
        Initializes an instance of AsyncBatchProcessor.

        Args:
            processor_fn (Callable[[List[T]], None]): The function to process the batches of items.
            max_batch_size (int, optional): The maximum size of each batch. Defaults to 100.
            min_batch_interval (float, optional): The minimum interval between processing batches. Defaults to 1.0.
        """
        self.processor_fn = processor_fn
        self.max_batch_size = max_batch_size
        self.min_batch_interval = min_batch_interval
        self.queue: Queue[T] = Queue()
        self.lock = Lock()
        self.stop_event = Event()  # Use an event to signal stopping
        self.processing_thread = Thread(target=self._process_batches)
        self.processing_thread.daemon = True
        self.processing_thread.start()
        atexit.register(self.wait_until_all_processed)  # Register cleanup function

    def enqueue(self, items: List[T]) -> None:
        """
        Enqueues a list of items to be processed.

        Args:
            items (List[T]): The items to be processed.
        """
        with self.lock:
            for item in items:
                self.queue.put(item)

    def _process_batches(self) -> None:
        """Internal method that continuously processes batches of items from the queue."""
        while True:
            current_batch: List[T] = []
            while not self.queue.empty() and len(current_batch) < self.max_batch_size:
                current_batch.append(self.queue.get())

            if current_batch:
                try:
                    self.processor_fn(current_batch)
                except requests.HTTPError as e:
                    if e.response.status_code == 413:
                        # 413: payload too large, don't raise just log
                        logger.error(f"Error processing batch: {e}")
                        if get_raise_on_captured_errors():
                            raise
                    else:
                        raise e

            if self.stop_event.is_set() and self.queue.empty():
                break

            # Unless we are stopping, sleep for a the min_batch_interval
            if not self.stop_event.is_set():
                time.sleep(self.min_batch_interval)

    def wait_until_all_processed(self) -> None:
        """Waits until all enqueued items have been processed."""
        self.stop_event.set()
        self.processing_thread.join()



"""Functions for inferring Weave Types from Python types."""

import collections
import types
import typing

import typing_extensions

from weave.legacy.weave import errors, graph

from weave.legacy.weave import weave_types


class TypedDictLike:
    __required_keys__: frozenset[str]


def is_typed_dict_like(t: type) -> typing_extensions.TypeGuard[TypedDictLike]:
    return hasattr(t, "__required_keys__")


def simple_python_type_to_type(py_type: type):  # type: ignore
    if isinstance(py_type, str):
        raise errors.WeaveTypeError(
            "Cannot yet detect Weave type from forward type references"
        )
    types = weave_types.instance_class_to_potential_type(py_type)
    if not types:
        return weave_types.UnknownType()
    return types[-1]  # last Type is most specific


def python_type_to_type(
    py_type: typing.Union[types.GenericAlias, type],
) -> weave_types.Type:
    if py_type == typing.Any:
        return weave_types.Any()
    elif isinstance(py_type, typing.TypeVar):
        if py_type.__bound__ is None:
            return weave_types.Any()
        else:
            return python_type_to_type(py_type.__bound__)
    elif isinstance(py_type, types.GenericAlias) or isinstance(
        py_type,
        typing._GenericAlias,  # type: ignore
    ):
        if py_type.__origin__ == typing.Literal:
            members = [
                weave_types.Const(weave_types.TypeRegistry.type_of(v), v)
                for v in py_type.__args__
            ]
            return weave_types.union(*members)
        args = [python_type_to_type(a) for a in py_type.__args__]
        if py_type.__origin__ == list or py_type.__origin__ == collections.abc.Sequence:
            return weave_types.List(*args)
        elif py_type.__origin__ == dict:
            # Special case, we return dict instead of TypedDict
            return weave_types.Dict(*args)
        elif py_type.__origin__ == typing.Union:
            return weave_types.UnionType(*args)
        elif py_type.__origin__ == graph.Node:
            return weave_types.Function({}, args[0])
        else:
            weave_type = simple_python_type_to_type(py_type.__origin__)
            if weave_type == weave_types.UnknownType():
                return weave_type
            return weave_type(*args)
    elif is_typed_dict_like(py_type):
        prop_types = {}
        not_required_keys = set()
        for (
            k,
            t,
        ) in py_type.__annotations__.items():
            if isinstance(t, typing.ForwardRef):
                # Its a ForwardRef if we use typing_extensions.TypedDict
                # which we have to when we want to use NotRequired.
                # But it can be immediately evaluated in the cases we
                # use it.
                t = t._evaluate(
                    {"NotRequired": typing_extensions.NotRequired}, None, frozenset()
                )
                if (
                    hasattr(t, "__origin__")
                    and t.__origin__ == typing_extensions.NotRequired
                ):
                    not_required_keys.add(k)
                    t = t.__args__[0]
            prop_types[k] = python_type_to_type(t)
        return weave_types.TypedDict(prop_types, not_required_keys=not_required_keys)
    weave_type = simple_python_type_to_type(py_type)
    if weave_type == weave_types.UnknownType():
        return weave_type
    try:
        return weave_type()
    except TypeError:
        raise errors.WeaveDefinitionError(
            "Can't instantatiate Weave Type %s without arguments. To fix: ensure all fields have defaults."
            % weave_type
        )



import typing
from dataclasses import dataclass
from urllib import parse

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import artifact_fs, decorator_type
from weave.legacy.weave.artifact_wandb import (
    WandbArtifact,
    WeaveWBLoggedArtifactURI,
)
from weave.legacy.weave.ops_domain.run_history import history_op_common
from weave.legacy.weave.wandb_interface import wandb_stream_table
from weave.legacy.weave.runfiles_wandb import WandbRunFiles, WeaveWBRunFilesURI


@dataclass
class RunPath:
    entity_name: str
    project_name: str
    run_name: str


# Avoiding name collision with PanelHistogram ('histogram') for now.
@decorator_type.type()
class WbHistogram:
    bins: list[float]
    values: list[int]


def _filesystem_artifact_file_from_artifact_path(artifact_path: str):
    uri = WeaveWBLoggedArtifactURI.parse(artifact_path)
    artifact = WandbArtifact(uri=uri, name=uri.name)
    file_path = uri.path
    if file_path is None:
        raise ValueError("Artifact path must contain a file path")
    return artifact_fs.FilesystemArtifactFile(artifact, file_path)


def _filesystem_runfiles_from_run_path(run_path: RunPath, file_path: str):
    uri = WeaveWBRunFilesURI.from_run_identifiers(
        run_path.entity_name,
        run_path.project_name,
        run_path.run_name,
    )
    runfiles = WandbRunFiles(name=uri.name, uri=uri)
    return runfiles.path_info(file_path)


def process_run_dict_obj(run_dict, run_path: typing.Optional[RunPath] = None):
    return {
        k: _process_run_dict_item(v, run_path)
        for k, v in run_dict.items()
        if k != "_wandb"
    }


def escape_artifact_path(artifact_path: str) -> str:
    prefix = "wandb-client-artifact://"
    if artifact_path.startswith(prefix):
        artifact_path = artifact_path[len(prefix) :]
        if ":" in artifact_path:
            name, version_path = artifact_path.split(":", 1)
            version, path = version_path.split("/", 1)
        else:
            version = None
            name, path = artifact_path.split("/", 1)
        path = parse.quote(path, safe="")
        version_string = f":{version}" if version is not None else ""
        artifact_path = f"{prefix}{name}{version_string}/{path}"
    return artifact_path


def _process_run_dict_item(val, run_path: typing.Optional[RunPath] = None):
    if isinstance(val, dict) and "_type" in val:
        if val["_type"] == "histogram":
            if "packedBins" in val:
                bins = []
                bin_min = val["packedBins"]["min"]
                for i in range(int(val["packedBins"]["count"])):
                    bins.append(bin_min)
                    bin_min += val["packedBins"]["size"]
            else:
                bins = val["bins"]
            return WbHistogram(
                bins=bins,
                values=val["values"],
            )
        if val["_type"] == "table-file":
            if "artifact_path" in val:
                artifact_path = escape_artifact_path(val["artifact_path"])
                return _filesystem_artifact_file_from_artifact_path(artifact_path)
            elif "path" in val and run_path is not None:
                return _filesystem_runfiles_from_run_path(run_path, val["path"])

        if val["_type"] in ["joined-table", "partitioned-table"]:
            return _filesystem_artifact_file_from_artifact_path(val["artifact_path"])

        if val["_type"] == "image-file" and run_path is not None:
            from weave.legacy.weave.ops_domain import ImageArtifactFileRef

            fs_artifact_file = _filesystem_runfiles_from_run_path(run_path, val["path"])
            return ImageArtifactFileRef(
                fs_artifact_file.artifact,
                fs_artifact_file.path,
                format=val["format"],
                width=val["width"],
                height=val["height"],
                sha256=val["sha256"],
                # boxes=val.get("boxes", {}),
                # masks=val.get("masks", {}),
            )
        if val["_type"] == "audio-file" and run_path is not None:
            from weave.legacy.weave.ops_domain import AudioArtifactFileRef

            fs_artifact_file = _filesystem_runfiles_from_run_path(run_path, val["path"])
            return AudioArtifactFileRef(  # type: ignore
                fs_artifact_file.artifact,
                fs_artifact_file.path,
                sha256=val["sha256"],
            )
        if val["_type"] == "bokeh-file" and run_path is not None:
            from weave.legacy.weave.ops_domain import BokehArtifactFileRef

            fs_artifact_file = _filesystem_runfiles_from_run_path(run_path, val["path"])
            return BokehArtifactFileRef(  # type: ignore
                fs_artifact_file.artifact,
                fs_artifact_file.path,
                sha256=val["sha256"],
            )
        if val["_type"] == "video-file" and run_path is not None:
            from weave.legacy.weave.ops_domain import VideoArtifactFileRef

            fs_artifact_file = _filesystem_runfiles_from_run_path(run_path, val["path"])
            return VideoArtifactFileRef(  # type: ignore
                fs_artifact_file.artifact,
                fs_artifact_file.path,
                sha256=val["sha256"],
            )
        if val["_type"] == "object3D-file" and run_path is not None:
            from weave.legacy.weave.ops_domain import Object3DArtifactFileRef

            fs_artifact_file = _filesystem_runfiles_from_run_path(run_path, val["path"])
            return Object3DArtifactFileRef(  # type: ignore
                fs_artifact_file.artifact,
                fs_artifact_file.path,
                sha256=val["sha256"],
            )
        if val["_type"] == "molecule-file" and run_path is not None:
            from weave.legacy.weave.ops_domain import MoleculeArtifactFileRef

            fs_artifact_file = _filesystem_runfiles_from_run_path(run_path, val["path"])
            return MoleculeArtifactFileRef(  # type: ignore
                fs_artifact_file.artifact,
                fs_artifact_file.path,
                sha256=val["sha256"],
            )
        if val["_type"] == "html-file" and run_path is not None:
            from weave.legacy.weave.ops_domain import HtmlArtifactFileRef

            fs_artifact_file = _filesystem_runfiles_from_run_path(run_path, val["path"])
            return HtmlArtifactFileRef(  # type: ignore
                fs_artifact_file.artifact,
                fs_artifact_file.path,
                sha256=val["sha256"],
            )

        if val["_type"] == "images/separated" and run_path is not None:
            from weave.legacy.weave.ops_domain import ImageArtifactFileRef

            image_list: list[ImageArtifactFileRef] = []

            for path in val["filenames"]:
                fs_artifact_file = _filesystem_runfiles_from_run_path(run_path, path)
                image_list.append(
                    ImageArtifactFileRef(
                        fs_artifact_file.artifact,
                        fs_artifact_file.path,
                        format=val["format"],
                        width=val["width"],
                        height=val["height"],
                        sha256=val.get("sha256", path),
                    )
                )

            return image_list
        if val["_type"] == "wb_trace_tree":
            from weave.legacy.weave.ops_domain.trace_tree import WBTraceTree

            return WBTraceTree(
                root_span_dumps=val.get("root_span_dumps"),  # type: ignore
                model_dict_dumps=val.get("model_dict_dumps"),
                model_hash=val.get("model_hash"),
            )

        if wandb_stream_table.is_weave_encoded_history_cell(val):
            return wandb_stream_table.from_weave_encoded_history_cell(val)

    return val


def process_run_dict_type(run_dict):
    return types.TypedDict(
        {
            k: _process_run_dict_item_type(v)
            for k, v in run_dict.items()
            if k != "_wandb"
        }
    )


def _process_run_dict_item_type(val):
    if isinstance(val, dict):
        type_count = {"type": val.get("_type", None)}
        type = history_op_common.history_key_type_count_to_weave_type(type_count)
        if type != types.UnknownType():
            return type
    return types.TypeRegistry.type_of(val)



import logging
import os
import pathlib
import shutil
import typing

import pytest
from flask.testing import FlaskClient

from weave.legacy.weave import client as client_legacy
from weave.legacy.weave import context_state, environment, io_service, serialize
from weave.legacy.weave.language_features.tagging.tag_store import (
    isolated_tagging_context,
)

from weave.legacy.weave import logs
# from .tests.wandb_system_tests_conftest import *

logs.configure_logger()

# Lazy mode was the default for a long time. Eager is now the default for the user API.
# A lot of tests are written to expect lazy mode, so just make lazy mode the default for
# tests.
context_state._eager_mode.set(False)

# A lot of tests rely on weave.legacy.weave.ops.* being in scope. Importing this here
# makes that work...

### Disable datadog engine tracing

def pytest_sessionstart(session):
    context_state.disable_analytics()


@pytest.fixture(autouse=True)
def pre_post_each_test(test_artifact_dir, caplog):
    # TODO: can't get this to work. I was trying to setup pytest log capture
    # to use our custom log stuff, so that it indents nested logs properly.
    caplog.handler.setFormatter(logging.Formatter(logs.default_log_format))
    # Tests rely on full cache mode right now.
    os.environ["WEAVE_CACHE_MODE"] = "full"
    os.environ["WEAVE_GQL_SCHEMA_PATH"] = str(
        pathlib.Path(__file__).parent.parent.parent / "wb_schema.gql"
    )
    try:
        shutil.rmtree(test_artifact_dir)
    except (FileNotFoundError, OSError):
        pass
    os.environ["WEAVE_LOCAL_ARTIFACT_DIR"] = test_artifact_dir
    with isolated_tagging_context():
        yield
    del os.environ["WEAVE_LOCAL_ARTIFACT_DIR"]




@pytest.fixture()
def fresh_server_logfile():
    def _clearlog():
        try:
            os.remove(logs.default_log_filename())
        except (OSError, FileNotFoundError) as e:
            pass

    _clearlog()
    yield
    _clearlog()

@pytest.fixture()
def eager_mode():
    with context_state.eager_execution():
        yield


@pytest.fixture()
def use_server_gql_schema():
    old_schema_path = environment.gql_schema_path()
    if old_schema_path is not None:
        del os.environ["WEAVE_GQL_SCHEMA_PATH"]
    yield
    if old_schema_path is not None:
        os.environ["WEAVE_GQL_SCHEMA_PATH"] = old_schema_path


class HttpServerTestClient:
    def __init__(self, flask_test_client: FlaskClient):
        """Constructor.

        Args:
            flask_test_client: A flask test client to use for sending requests to the test application
        """
        self.flask_test_client = flask_test_client
        self.execute_endpoint = "/__weave/execute"

    def execute(
        self,
        nodes,
        headers: typing.Optional[dict[str, typing.Any]] = None,
        no_cache=False,
    ):
        _headers: dict[str, typing.Any] = {}
        if headers is not None:
            _headers = headers

        serialized = serialize.serialize(nodes)
        r = self.flask_test_client.post(
            self.execute_endpoint,
            json={"graphs": serialized},
            headers=_headers,
        )

        return r.json["data"]


@pytest.fixture()
def io_server_factory():
    original_server = io_service.SERVER

    def factory(process=False):
        server = io_service.Server(process=process)
        server.start()
        io_service.SERVER = server
        return server

    yield factory

    if io_service.SERVER and io_service.SERVER is not original_server:
        io_service.SERVER.shutdown()

    io_service.SERVER = original_server


@pytest.fixture()
def http_server_test_client(app):
    from .. import weave_server

    app = weave_server.make_app()
    flask_client = app.test_client()
    return HttpServerTestClient(flask_client)


@pytest.fixture()
def weave_test_client(http_server_test_client):
    return client_legacy.Client(http_server_test_client)




@pytest.fixture()
def ref_tracking():
    with context_state.ref_tracking(True):
        yield





import pytest

import weave

types = weave.types.get_type_classes()


@pytest.mark.parametrize("type_name, type_cls", [(t.name, t) for t in types])
def test_const_assignment(type_name, type_cls):
    from weave.legacy.weave.ops_domain import wb_domain_types as wdt

    params = []
    # Quick list of types that don't work with this parametrization
    if type_name in ["const", "invalid", "union", "function", "unknown"]:
        return

    # Special types that need extra params
    if type_name == "list":
        params = [weave.types.String()]
    if type_name == "typedDict":
        params = [{"col": weave.types.String()}]
    if type_name == "dict":
        params = [weave.types.String(), weave.types.String()]
    if type_name == "dataframe":
        params = [weave.types.TypedDict({"col": weave.types.String()})]
    if type_name == "tagged":
        params = [
            weave.types.TypedDict({"col": weave.types.String()}),
            weave.types.String(),
        ]
    if type_name == "PartialObject":
        params = [
            type(wdt.RunType),
            weave.types.TypedDict({"col": weave.types.String()}),
        ]

    cls_type = type_cls(*params)
    const_type = weave.types.Const(
        cls_type, None
    )  # technically `None` is invalid, but we're just testing the assignment

    # Validate that you can assign a const type to a general type
    assert cls_type.assign_type(const_type)

    # strict Const assignability is disabled. See weave_types.py and
    # test_plot_constants_assign
    # Validate that you cannot assign a general type to a const type
    # assert not const_type.assign_type(cls_type)

    # Validate that unions are automatically unwrapped
    union_type = weave.types.UnionType(cls_type, const_type)
    assert union_type.assign_type(const_type)
    assert union_type.assign_type(cls_type)
    assert cls_type.assign_type(union_type)

    # strict Const assignability is disabled. See weave_types.py and
    # test_plot_constants_assign
    # assert not const_type.assign_type(union_type)



import wandb

import weave
from weave.legacy.weave import compile


# Example of end to end integration test
def test_run_logging(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    run.log({"a": 1})
    run.finish()

    summary_node = (
        weave.legacy.weave.ops.project(run.entity, run.project).run(run.id).summary()["a"]
    )
    summary = weave.use(summary_node)

    assert summary == 1

    is_none_node = weave.legacy.weave.ops.project(run.entity, run.project).isNone()

    assert weave.use(is_none_node) == False

    is_none_node = weave.legacy.weave.ops.project(
        run.entity, "project_does_not_exist"
    ).isNone()

    assert weave.use(is_none_node) == True


# Test each of the auth strategies
def test_basic_publish_user_by_api_key_in_context(user_by_api_key_in_context):
    return _test_basic_publish(user_by_api_key_in_context)


def test_basic_publish_user_by_api_key_in_env(user_by_api_key_in_env):
    return _test_basic_publish(user_by_api_key_in_env)


def test_basic_publish_user_by_api_key_netrc(user_by_api_key_netrc):
    return _test_basic_publish(user_by_api_key_netrc)


# As of this writing, I don't have a good way to bootstrap cookies
# def test_basic_publish_user_by_http_headers_in_context(user_by_http_headers_in_context):
#     return _test_basic_publish(user_by_http_headers_in_context)


# def test_basic_publish_user_by_http_headers_in_env(user_by_http_headers_in_env):
#     return _test_basic_publish(user_by_http_headers_in_env)


def _test_basic_publish(user_fixture):
    ref = weave.storage.publish([1, 2, 3], "weave/list")
    # Getting the ref for `a` is not a final API. Expect
    # that changes to the publish flow will bread this test
    # and you might need to update how you get the ref.
    uri = ref.uri
    assert (
        uri
        == f"wandb-artifact:///{user_fixture.username}/weave/list:0cdf3358dc939f961ca9/obj"
    )
    assert weave.legacy.weave.ref_base.Ref.from_str(uri).get() == [1, 2, 3]


# Example of end to end integration test
def test_run_histories(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    run.log({"a": 1})
    run.finish()

    run = wandb.init(project="project_exists")
    run.log({"a": 2})
    run.finish()

    history_node = (
        weave.legacy.weave.ops.project(run.entity, run.project).runs().history().concat()["a"]
    )
    history = weave.use(history_node)

    # Runs return in reverse chronological order
    assert history == [2, 1]


# Example of end to end integration test
def test_run_history_count(user_by_api_key_in_env, cache_mode_minimal):
    run = wandb.init(project="project_exists")
    run.log({"a": 1})
    run.log({"a": 2})
    run.finish()

    run_node = weave.legacy.weave.ops.project(run.entity, run.project).run(run.id)
    h_count_node = run_node.history().count()
    history_count = weave.use(h_count_node)
    assert history_count == 2

    compiled = compile.compile([h_count_node])
    assert compiled[0].from_op.name == "run-historyLineCount"

    run_history_lines = weave.use(run_node.historyLineCount())
    assert run_history_lines == 2



import atexit
import contextlib
import datetime
import json
import logging
import os
import queue
import random
import shutil
import threading
import typing
import uuid

from wandb.sdk.lib.ipython import _get_python_type
from wandb.sdk.lib.paths import LogicalPath
from wandb.sdk.lib.printer import get_printer

from weave.legacy.weave import weave_types
from weave.legacy.weave import (
    storage, 
    errors, 
    artifact_base,
    box,
    environment,
    file_util,
    graph,
    runfiles_wandb,
    wandb_api,
)
from weave.legacy.weave.core_types.stream_table_type import StreamTableType
from weave.legacy.weave.wandb_interface.wandb_lite_run import InMemoryLazyLiteRun

if typing.TYPE_CHECKING:
    from wandb.sdk.internal.file_pusher import FilePusher


# Shawn recommended we only encode leafs, but in my testing, nested structures
# are not handled as well in in gorilla and we can do better using just weave.
# Uncomment the below to use gorilla for nested structures.
TRUST_GORILLA_FOR_NESTED_STRUCTURES = True

# Weave types are parametrized, but gorilla expects just simple strings. We could
# send the top-level string over the wire, but this fails to encode type specifics
# and therefore loses information. With this flag, we instead stringify the JSON type
# and send that over the wire. This is a bit of a hack, but it works.
ENCODE_ENTIRE_TYPE = True
TYPE_ENCODE_PREFIX = "_wt_::"

ROW_TYPE = typing.Union[typing.Mapping, list[typing.Mapping]]


class WandbLiveRunFiles(runfiles_wandb.WandbRunFiles):
    _file_pusher: typing.Optional["FilePusher"] = None
    _temp_dir: typing.Optional[str] = None

    def set_file_pusher(self, pusher: "FilePusher") -> None:
        # It is the responsibility of the caller to ensure this file pusher is
        # correctly associated with the corresponding run.
        self._file_pusher = pusher

    def temp_dir(self) -> str:
        if self._temp_dir is None:
            rand_part = "".join(random.choice("0123456789ABCDEF") for _ in range(16))
            self._temp_dir = os.path.join(
                runfiles_wandb.wandb_run_dir(), "upload_cache", f"tmp_{rand_part}"
            )
            os.makedirs(self._temp_dir, exist_ok=True)
        return self._temp_dir

    def cleanup(self) -> None:
        if self._temp_dir is not None:
            shutil.rmtree(self._temp_dir)
            self._temp_dir = None

    def __del__(self) -> None:
        self.cleanup()

    @contextlib.contextmanager
    def new_file(
        self, path: str, binary: bool = False
    ) -> typing.Generator[typing.IO, None, None]:
        if self._file_pusher is None:
            raise ValueError(
                "WandbLiveRunFiles must be associated with a file pusher to use new_file"
            )

        dir_path = self.temp_dir()
        file_path = os.path.join(dir_path, path)
        with file_util.safe_open(file_path, "wb" if binary else "w") as file:
            yield file
            self._file_pusher.file_changed(LogicalPath(path), file_path)


class _StreamTableSync:
    _lite_run: InMemoryLazyLiteRun
    _table_name: str
    _project_name: str
    _entity_name: str

    _artifact: WandbLiveRunFiles

    _weave_stream_table: StreamTableType
    _weave_stream_table_ref: artifact_base.ArtifactRef

    _client_id: str

    def __init__(
        self,
        table_name: str,
        *,
        project_name: typing.Optional[str] = None,
        entity_name: typing.Optional[str] = None,
        _disable_async_file_stream: bool = False,
    ):
        self._client_id = str(uuid.uuid1())
        splits = table_name.split("/")
        if len(splits) == 1:
            pass
        elif len(splits) == 2:
            if project_name is not None:
                raise ValueError(
                    f"Cannot specify project_name and table_name with '/' in it: {table_name}"
                )
            project_name = splits[0]
            table_name = splits[1]
        elif len(splits) == 3:
            if project_name is not None or entity_name is not None:
                raise ValueError(
                    f"Cannot specify project_name or entity_name and table_name with 2 '/'s in it: {table_name}"
                )
            entity_name = splits[0]
            project_name = splits[1]
            table_name = splits[2]

        if entity_name is None or entity_name == "":
            # Use user's default entity
            api = wandb_api.get_wandb_api_sync()
            entity_name = api.default_entity_name()
            # get default entity
            if entity_name is None or entity_name == "":
                raise ValueError(
                    "Entity not specified and no default entity found. Please specify entity_name or set default entity with `wandb login --entity <entity_name>`"
                )

        if project_name is None or project_name == "":
            raise ValueError("Must specify project_name")
        elif table_name is None or table_name == "":
            raise ValueError("Must specify table_name")

        self._lite_run = InMemoryLazyLiteRun(
            entity_name,
            project_name,
            table_name,
            group="weave_stream_tables",
            _hide_in_wb=True,
            _use_async_file_stream=not _disable_async_file_stream,
        )
        self._table_name = self._lite_run._run_name
        self._project_name = self._lite_run._project_name
        self._entity_name = self._lite_run._entity_name
        self._ensure_remote_initialized()
        atexit.register(self._at_exit)

    def _ensure_remote_initialized(self) -> StreamTableType:
        self._lite_run.ensure_run()
        print_url = False
        if not hasattr(self, "_weave_stream_table"):
            print_url = True
            self._weave_stream_table = StreamTableType(  # type: ignore
                table_name=self._table_name,
                project_name=self._project_name,
                entity_name=self._entity_name,
            )
            self._weave_stream_table_ref = storage._direct_publish(
                self._weave_stream_table,
                name=self._table_name,
                wb_project_name=self._project_name,
                wb_entity_name=self._entity_name,
                _lite_run=self._lite_run,
            )
        if not hasattr(self, "_artifact"):
            uri = runfiles_wandb.WeaveWBRunFilesURI.from_run_identifiers(
                self._entity_name,
                self._project_name,
                self._table_name,
            )
            self._artifact = WandbLiveRunFiles(name=uri.name, uri=uri)
            self._artifact.set_file_pusher(self._lite_run.pusher)
        if print_url:
            base_url = environment.weave_server_url()
            url = f"{base_url}/browse/wandb/{self._entity_name}/{self._project_name}/table/{self._table_name}"
            printer = get_printer(_get_python_type() != "python")
            # printer.display(f'{printer.emoji("star")} View data at {printer.link(url)}')
        return self._weave_stream_table

    def log(self, row_or_rows: ROW_TYPE) -> None:
        if isinstance(row_or_rows, dict):
            row_or_rows = [row_or_rows]

        for row in row_or_rows:
            self._log_row(row)

    def rows(self) -> graph.Node:
        from weave.legacy.weave.ops_domain import stream_table_ops
        from weave.legacy.weave.ops_primitives import weave_api

        if self._weave_stream_table_ref is None:
            raise errors.WeaveInternalError("ref is None after ensure")
        return stream_table_ops.rows(
            weave_api.get(str(self._weave_stream_table_ref.uri))
        )

    def _ipython_display_(self) -> graph.Node:
        from weave.legacy.weave import show

        return show.show(self.rows())

    def _log_row(self, row: typing.Mapping) -> None:
        row_copy = {**row}
        row_copy["_client_id"] = self._client_id
        if "timestamp" not in row_copy:
            row_copy["timestamp"] = datetime.datetime.now()
        payload = row_to_weave(row_copy, self._artifact)
        self._lite_run.log(payload)

    def finish(self) -> None:
        if self._lite_run:
            self._lite_run.finish()
        if self._artifact:
            self._artifact.cleanup()

    def __del__(self) -> None:
        self.finish()

    def _at_exit(self) -> None:
        self.finish()


class StreamTable(_StreamTableSync):
    MAX_UNSAVED_SECONDS = 2

    def __init__(
        self,
        table_name: str,
        *,
        project_name: typing.Optional[str] = None,
        entity_name: typing.Optional[str] = None,
        _disable_async_file_stream: bool = False,
    ):
        super().__init__(
            table_name=table_name,
            project_name=project_name,
            entity_name=entity_name,
            _disable_async_file_stream=_disable_async_file_stream,
        )

        self.queue: queue.Queue = queue.Queue()
        self._lock = threading.Lock()
        self._join_event = threading.Event()
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True
        self._thread.start()

    def log(self, row_or_rows: ROW_TYPE) -> None:
        # weave_types.TypeRegistry.type_of(row_or_rows)
        self.queue.put(row_or_rows)

    def _flush(self) -> None:
        with self._lock:
            for log_payload in self._iterate_queue():
                super().log(log_payload)

    def _iterate_queue(
        self,
    ) -> typing.Generator[ROW_TYPE, None, None]:
        while True:
            try:
                record = self.queue.get_nowait()
            except queue.Empty:
                break
            else:
                yield record
                self.queue.task_done()

    def _thread_body(self) -> None:
        join_requested = False
        while not join_requested:
            join_requested = self._join_event.wait(self.MAX_UNSAVED_SECONDS)
            self._flush()

    # Override methods of _StreamTableSync
    def finish(self) -> None:
        if hasattr(self, "_thread"):
            self._join_event.set()
            self._thread.join()
            with self._lock:
                super().finish()


def maybe_history_type_to_weave_type(tc_type: str) -> typing.Optional[weave_types.Type]:
    if tc_type.startswith(TYPE_ENCODE_PREFIX):
        w_type = json.loads(tc_type[len(TYPE_ENCODE_PREFIX) :])
        try:
            return weave_types.TypeRegistry.type_from_dict(w_type)
        except Exception as e:
            logging.warning(
                f"StreamTable Type Error: Found type for {tc_type}, but blind construction failed: {e}",
            )
    else:
        possible_type = weave_types.type_name_to_type(tc_type)
        if possible_type is not None:
            try:
                return possible_type()
            except Exception as e:
                logging.warning(
                    f"StreamTable Type Error: Found type for {tc_type}, but blind construction failed: {e}",
                )
    return None


def is_weave_encoded_history_cell(cell: dict) -> bool:
    return "_val" in cell and (
        "_weave_type" in cell
        or (cell.get("_type") != None and cell["_type"].startswith(TYPE_ENCODE_PREFIX))
    )


def from_weave_encoded_history_cell(cell: dict) -> typing.Any:
    if not is_weave_encoded_history_cell(cell):
        raise ValueError(f"Expected weave encoded history cell, got {cell}")
    if "_weave_type" in cell:
        weave_type = cell["_weave_type"]
    elif cell["_type"].startswith(TYPE_ENCODE_PREFIX):
        weave_type = json.loads(cell["_type"][len(TYPE_ENCODE_PREFIX) :])
    else:
        raise ValueError(f"Expected weave encoded history cell, got {cell}")
    weave_json = {
        "_type": weave_type,
        "_val": cell["_val"],
    }
    return storage.from_python(weave_json)


def row_to_weave(
    row: typing.Dict[str, typing.Any], artifact: WandbLiveRunFiles
) -> typing.Dict[str, typing.Any]:
    return {key: obj_to_weave(value, artifact) for key, value in row.items()}


def obj_to_weave(obj: typing.Any, artifact: WandbLiveRunFiles) -> typing.Any:
    def recurse(obj: typing.Any) -> typing.Any:
        return obj_to_weave(obj, artifact)

    # all primitives
    obj = box.unbox(obj)
    if isinstance(obj, (int, float, str, bool, type(None))):
        return obj
    else:
        if TRUST_GORILLA_FOR_NESTED_STRUCTURES:
            if isinstance(obj, dict):
                return {key: recurse(value) for key, value in obj.items()}
            elif isinstance(obj, list):
                return [recurse(value) for value in obj]
            elif isinstance(obj, tuple):
                return [recurse(value) for value in obj]
            elif isinstance(obj, set):
                return [recurse(value) for value in obj]
            elif isinstance(obj, frozenset):
                return [recurse(value) for value in obj]
            else:
                return leaf_to_weave(obj, artifact)
        else:
            return leaf_to_weave(obj, artifact)


def w_type_to_type_name(w_type: typing.Union[str, dict]) -> str:
    if isinstance(w_type, str):
        return w_type
    if ENCODE_ENTIRE_TYPE:
        return TYPE_ENCODE_PREFIX + json.dumps(w_type)
    else:
        return w_type["type"]


def leaf_to_weave(leaf: typing.Any, artifact: WandbLiveRunFiles) -> typing.Any:
    def ref_persister_artifact(
        type: weave_types.Type, refs: typing.Iterable[artifact_base.ArtifactRef]
    ) -> artifact_base.Artifact:
        # Save all the reffed objects into the new artifact.
        for mem_ref in refs:
            if mem_ref.path is not None and mem_ref._type is not None:
                # Hack: add a random salt to the end (i really want content addressing here)
                # but this is a quick fix to avoid collisions
                path = mem_ref.path + "-" + str(uuid.uuid4())
                artifact.set(path, mem_ref._type, mem_ref._obj)
        return artifact

    res = storage.to_python(leaf, None, ref_persister_artifact)

    w_type = res["_type"]
    type_name = w_type_to_type_name(w_type)

    if ENCODE_ENTIRE_TYPE:
        return {"_type": type_name, "_val": res["_val"]}
    else:
        return {
            "_type": type_name,
            "_weave_type": w_type,
            "_val": res["_val"],
        }



"""
This file contains a single exported function `process_opdef_resolve_fn`. Technically,
this could have been implemented in `op_def.py`, but we want to keep all the tagging-related
logic in one place. Therefore, in op_def.py we import this function and call it to post-process
the result of the op_def's resolve_fn.
"""

import typing

import pyarrow as pa
import typing_extensions
from pyarrow import compute as pc

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import box
from weave.legacy.weave.arrow.arrow_tags import awl_add_arrow_tags
from weave.legacy.weave.arrow.list_ import ArrowWeaveList
from weave.legacy.weave.language_features.tagging import tag_store
from weave.legacy.weave.language_features.tagging.opdef_util import (
    get_first_arg,
    should_flow_tags,
    should_tag_op_def_outputs,
)
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType

if typing.TYPE_CHECKING:
    from weave.legacy.weave import op_def as OpDef


def _is_tagged_value(val: types.Type) -> typing_extensions.TypeGuard[TaggedValueType]:
    return isinstance(val, TaggedValueType)


def _is_optional_tagged_value(
    val: types.Type,
) -> typing_extensions.TypeGuard[types.UnionType]:
    return types.is_optional(val) and isinstance(types.non_none(val), TaggedValueType)


def _strip_tags(val: typing.Any) -> typing.Any:
    from weave.legacy.weave.ops_arrow import ArrowWeaveList

    if isinstance(val, ArrowWeaveList):
        if _is_tagged_value(val.object_type):
            return ArrowWeaveList(
                val._arrow_data.field("_value"),
                val.object_type.value,
                val._artifact,
            )

        elif _is_optional_tagged_value(val.object_type):
            return ArrowWeaveList(
                val._arrow_data.field("_value"),
                typing.cast(TaggedValueType, types.non_none(val.object_type)).value,
                val._artifact,
            )
    return val


# here we strip the tags from each element of an arrow weave list and create a
# new arrow weave list that contains just the untagged values. We then pass that
# to the op resolver. On we get the result back, we re-apply the tags we stripped
# previously
def propagate_arrow_tags(
    op_def: "OpDef.OpDef",
    resolve_fn: typing.Callable,
    args: list[typing.Any],
    kwargs: dict[str, typing.Any],
) -> typing.Any:
    tag_type: typing.Optional[types.Type]

    _, first_arg_val = get_first_arg(op_def, args, kwargs)
    first_arg_val = typing.cast(ArrowWeaveList, first_arg_val)

    is_optional_tagged = False
    if _is_tagged_value(first_arg_val.object_type):
        first_arg_tags = first_arg_val._arrow_data.field("_tag")
        tag_type = first_arg_val.object_type.tag
    elif _is_optional_tagged_value(first_arg_val.object_type):
        first_arg_tags = first_arg_val._arrow_data.field("_tag")
        tag_type = typing.cast(
            TaggedValueType, types.non_none(first_arg_val.object_type)
        ).tag
        is_optional_tagged = True
    else:
        first_arg_tags = None
        tag_type = None

    tag_stripped_args: list[typing.Any] = []
    tag_stripped_kwargs: dict[str, typing.Any] = {}

    for arg in args:
        tag_stripped_args.append(_strip_tags(arg))

    for key, val in kwargs.items():
        tag_stripped_kwargs[key] = _strip_tags(val)

    res = resolve_fn(*tag_stripped_args, **tag_stripped_kwargs)

    # rewrap tags
    if first_arg_tags and tag_type:
        res = awl_add_arrow_tags(
            res,
            first_arg_tags,
            tag_type,
        )

        if is_optional_tagged:
            mask = pc.invert(pc.is_valid(first_arg_val._arrow_data))
            new_arrow_data = pa.StructArray.from_arrays(
                [res._arrow_data.field("_tag"), res._arrow_data.field("_value")],
                ["_tag", "_value"],
                mask=mask,
            )

            res = ArrowWeaveList(
                new_arrow_data,
                types.optional(res.object_type),
                res._artifact,
            )

    return res


def flow_tags(
    first_input: typing.Any,
    output: typing.Any,
    give_precedence_to_existing_tags: bool = False,
) -> typing.Any:
    if tag_store.is_tagged(first_input):
        tag_dict = tag_store.get_tags(first_input)
        if isinstance(output, types.Type):
            if len(tag_dict.keys()) == 0:
                return output
            tag_type = types.TypeRegistry.type_of(tag_dict)
            assert isinstance(tag_type, types.TypedDict)
            return TaggedValueType(tag_type, output)
        return tag_store.add_tags(
            output,
            tag_dict,
            give_precedence_to_existing_tags=give_precedence_to_existing_tags,
        )
    return output


# This function is responsible for post-processing the results of a resolve_fn.
# Specifically, it will take one of 3 actions:
# 1. If `_should_tag_op_def_outputs` is true, then it will tag the output with the input.
# 2. Else If `_should_flow_tags`, then it will flow the tags from the input to the output.
# 3. Else, it will just return the output.
def process_opdef_resolve_fn(
    op_def: "OpDef.OpDef",
    resolve_fn: typing.Callable,
    args: list[typing.Any],
    kwargs: dict[str, typing.Any],
) -> typing.Any:
    if op_def.op_def_is_auto_tag_handling_arrow_op():
        res = propagate_arrow_tags(op_def, resolve_fn, args, kwargs)

        # TODO(DG): implement this for Table, ChunkedArray, etc.
        if isinstance(res._arrow_data, pa.Array) and res._arrow_data.null_count > 0:
            res.object_type = types.optional(res.object_type)

    else:
        res = resolve_fn(*args, **kwargs)

    res = box.box(res)
    if should_tag_op_def_outputs(op_def):
        key, val = get_first_arg(op_def, args, kwargs)
        result_is_identity = id(val) == id(res)
        if not result_is_identity:
            tag_dict = {key: val}
            if isinstance(res, types.Type):
                tag_type = types.TypeRegistry.type_of(tag_dict)
                assert isinstance(tag_type, types.TypedDict)
                return TaggedValueType(tag_type, res)
            return tag_store.add_tags(res, tag_dict)
    elif should_flow_tags(op_def):
        key, val = get_first_arg(op_def, args, kwargs)
        return flow_tags(val, res, give_precedence_to_existing_tags=True)
    return res



# "Stitch" a graph together, by traversing tag and literal construction and access.
#
# The resulting stitched graph is appropriate for optimizations like projection
# pushdown and predicate pushdown.
#
# Weave DAGs can contain ops that change the shape of values in ways that are entirely
# recoverable from the DAG. For example, you can use the dict_ op to create a dict
# that has values at known dict keys. The resulting dict may be passed around, and
# eventually have its keys accessed by getitem. This library attaches ops that operate
# on a downstream getitem back to the original item that was placed in the dict at the
# dict_ call.
#
# It also enters sub-DAGs, like the argument to map, and stitches encountered ops
# back to the map's input node.
#
# This implementation MUST be comprehensive. IE it must correctly get all downstream
# ops that should be stitched to a given node, or denote that its not possible. Optimizers
# need accurate information to optimize!
#
# Not yet implemented: list literals, execute() traversal, and probably a bunch more!

import dataclasses
import typing

from weave.legacy.weave import graph, op_def, registry_mem
from weave.legacy.weave.language_features.tagging import opdef_util

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from . import _dict_utils


@dataclasses.dataclass
class OpCall:
    node: graph.OutputNode
    input_dict: dict[str, "ObjectRecorder"]
    output: "ObjectRecorder"

    @property
    def inputs(self) -> list["ObjectRecorder"]:
        return list(self.input_dict.values())


@dataclasses.dataclass()
class ObjectRecorder:
    node: graph.Node
    tags: dict[str, "ObjectRecorder"] = dataclasses.field(default_factory=dict)
    val: typing.Optional[typing.Any] = None
    calls: list[OpCall] = dataclasses.field(default_factory=list)

    def call_node(
        self, node: graph.OutputNode, input_dict: dict[str, "ObjectRecorder"]
    ) -> "ObjectRecorder":
        output = ObjectRecorder(node)
        self.calls.append(OpCall(node, input_dict, output))
        return output


class LiteralDictObjectRecorder(ObjectRecorder):
    val: dict[str, ObjectRecorder]


class LiteralListObjectRecorder(ObjectRecorder):
    val: list[ObjectRecorder]


def node_shortname(node: graph.Node) -> str:
    if isinstance(node, graph.OutputNode):
        return node.from_op.name
    elif isinstance(node, graph.ConstNode):
        return f"C[{str(node.val)[:10]}]"
    elif isinstance(node, graph.VarNode):
        return f"V[{node.name}]"
    else:
        raise errors.WeaveInternalError(f"Unknown node type {type(node)}")


@dataclasses.dataclass
class StitchedGraph:
    _node_map: typing.Dict[graph.Node, ObjectRecorder]

    def has_result(self, node: graph.Node) -> bool:
        return node in self._node_map

    def get_result(self, node: graph.Node) -> ObjectRecorder:
        return self._node_map[node]

    def add_result(self, node: graph.Node, result: ObjectRecorder) -> None:
        self._node_map[node] = result

    def print_debug_summary(self) -> None:
        res = ""
        node_names = {
            orig_node: f"<{i}-{node_shortname(orig_node)}>"
            for i, (orig_node, recorder) in enumerate(self._node_map.items())
        }
        res += "\n" + "StitchedGraph Summary:"
        res += "\n" + "  Nodes:"
        for curr_node, curr_node_name in node_names.items():
            if isinstance(curr_node, graph.OutputNode):
                res += (
                    "\n"
                    + f"  * {curr_node_name}({','.join([node_names[input_node] for input_node in  curr_node.from_op.inputs.values()])})"
                )
            else:
                res += "\n" + f"  * {curr_node_name}"
            recorder = self._node_map[curr_node]
            if recorder.calls:
                res += (
                    "\n"
                    + f"    calls: {','.join([node_names[call.node] for call in recorder.calls])}"
                )
            if recorder.tags:
                res += (
                    "\n"
                    + f"    tags: {','.join([str((tag_name, node_names[tag_recorder.node])) for tag_name, tag_recorder in recorder.tags.items()])}"
                )
        print(res)


def stitch(
    leaf_nodes: list[graph.Node],
    var_values: typing.Optional[dict[str, ObjectRecorder]] = None,
    stitched_graph: typing.Optional[StitchedGraph] = None,
    on_error: graph.OnErrorFnType = None,
) -> StitchedGraph:
    """Given a list of leaf nodes, stitch the graph together."""
    if stitched_graph is None:
        sg = StitchedGraph({})
    else:
        sg = stitched_graph

    def handle_node(node: graph.Node) -> graph.Node:
        if sg.has_result(node):
            return node
        if isinstance(node, graph.OutputNode):
            input_dict = {k: sg.get_result(v) for k, v in node.from_op.inputs.items()}
            sg.add_result(node, stitch_node(node, input_dict, sg))
        elif isinstance(node, graph.ConstNode):
            # Note from Tim: I believe this block of code should be correct (ie. stitching inside
            # of static lambdas). I wrote it while implementing a fix but it ended up not being needed.
            # Stitch is a particularly touchy part of the code base. So i would rather leave this
            # code commented out for now, but not delete it. My gut tells me that any static lambda
            # will be re-executed if needed and therefore re-stitched. And furthermore, any caller
            # of stitch right now (eg. gql compile) should explicitly NOT mutate static lambdas. So
            # it might be a foot-gun to allow this code to run, even if it is correct. Let's find a
            # use case for this code before we uncomment it.
            # is_static_lambda = (
            #     isinstance(node.type, types.Function)
            #     and len(node.type.input_types) == 0
            # )
            # if is_static_lambda:
            #     sg.add_result(node, subgraph_stitch(node.val, {}, sg))
            sg.add_result(node, ObjectRecorder(node, val=node.val))
        elif isinstance(node, graph.VarNode):
            if var_values and node.name in var_values:
                sg.add_result(node, var_values[node.name])
            else:
                # Originally, we would raise an error here, but we now allow
                # unbound vars to be in the graph. This makes sense to me (Tim)
                # since we could be stitching the compilation step of a
                # subgraph, but Shawn has pointed out that this case could be
                # removed if we stitched the entire graph in 1 pass. For now, we
                # assign an empty recorder, but this might be able to be
                # optimized out in the future.
                sg.add_result(node, ObjectRecorder(node))
                # raise errors.WeaveInternalError(
                #     "Encountered var %s but var_values not provided" % node.name
                # )
        else:
            raise errors.WeaveInternalError("Unexpected node type")
        return node

    graph.map_nodes_top_level(leaf_nodes, handle_node, on_error)

    return sg


def subgraph_stitch(
    function_node: graph.Node, args: dict[str, ObjectRecorder], sg: StitchedGraph
) -> ObjectRecorder:
    result_graph = stitch([function_node], args, stitched_graph=sg)
    return result_graph.get_result(function_node)


def is_root_op(op: op_def.OpDef) -> bool:
    return (
        op.name == "root-project"
        or op.name == "get"
        or op.name == "getReturnType"
        or op.name == "render_table_runs2"
        or op.name == "project-runs2"
    )


def is_mapped_get_tag_op(op: op_def.OpDef) -> bool:
    if op.derived_from and op.derived_from.derived_ops.get("mapped") == op:
        return is_get_tag_op(op.derived_from)
    return False


def is_get_tag_op(op: op_def.OpDef) -> bool:
    return op._gets_tag_by_name != None


def get_tag_name_from_tag_getter_op(op: op_def.OpDef) -> str:
    assert (
        op._gets_tag_by_name != None
    ), "Caller should verify that this is a tag getter op"
    return op._gets_tag_by_name  # type: ignore


def get_tag_name_from_mapped_tag_getter_op(op: op_def.OpDef) -> str:
    assert (
        op.derived_from != None
    ), "Caller should verify that this is a mapped tag getter op"
    return get_tag_name_from_tag_getter_op(op.derived_from)  # type: ignore


def stitch_node_inner(
    node: graph.OutputNode, input_dict: dict[str, ObjectRecorder], sg: StitchedGraph
) -> ObjectRecorder:
    """
    It is the responsibility of the `stitch_node` function to do do two things:
    1. Ensure that the ObjectRecorder for the node is created (if not existing) and return it
    2. Ensure that all input `ObjectRecords` are connected to the ObjectRecorder in #1
        - These are two-way connections:
                - The ObjectRecorder in #1 has a list of inputs via `input_dict`
                - The ObjectRecorders in #2 have a list of outputs via `calls`

        Importantly, nodes which are `lambdas` (ex map). Need to traverse the inner function. This
        is done by calling `subgraph_stitch` and passing the current StitchedGraph. We need to pass
        the stitch graph so that the `subgraph_stitch` can merge the results of the inner function.
    """
    op = registry_mem.memory_registry.get_op(node.from_op.name)
    inputs = list(input_dict.values())
    if is_get_tag_op(op):
        tag_name = get_tag_name_from_tag_getter_op(op)
        return inputs[0].tags[tag_name]
    elif is_mapped_get_tag_op(op):
        tag_name = get_tag_name_from_mapped_tag_getter_op(op)
        return inputs[0].tags[tag_name]
    elif node.from_op.name.endswith("createIndexCheckpointTag"):
        inputs[0].tags["indexCheckpoint"] = ObjectRecorder(node)
        return inputs[0]
    elif node.from_op.name == "dict":
        return LiteralDictObjectRecorder(node, val=input_dict)
    elif node.from_op.name.endswith("__getitem__"):
        return inputs[0]
    elif node.from_op.name == "list":
        # Merge element tags together and place them on the outer list.
        # This is overly aggressive, but it means we don't need to provide
        # special handling for concat and other structural list ops for
        # now.
        tags: dict[str, ObjectRecorder] = {}
        for input in input_dict.values():
            tags.update(input.tags)
        res = LiteralListObjectRecorder(node, tags=tags, val=list(input_dict.values()))
        op_call = OpCall(node, input_dict, res)
        for input in input_dict.values():
            input.calls.append(op_call)
        return res
    elif node.from_op.name.endswith("pick"):
        if isinstance(node.from_op.inputs["key"], graph.ConstNode):
            path = _dict_utils.split_escaped_string(node.from_op.inputs["key"].val)
            if len(path) == 0:
                return ObjectRecorder(node, inputs[0].tags)
            key = _dict_utils.unescape_dots(path[0])
            if (
                isinstance(inputs[0], LiteralDictObjectRecorder)
                and key in inputs[0].val
            ):
                val = inputs[0].val.get(key)
                if val is not None:
                    if len(path) == 1:
                        return val
                    # If there are further path elements, recursively stitch.
                    sub_key = ".".join(path[1:])
                    # We manually construct a typedDict-pick op with the next sub-key
                    new_node = graph.OutputNode(
                        node.type,
                        "typedDict-pick",
                        {
                            # the first argument is unused by stitch_node_inner,
                            # so just set it to None
                            "arr": graph.ConstNode(types.NoneType(), None),
                            "key": graph.ConstNode(types.String(), sub_key),
                        },
                    )
                    # Make a new ObjectRecorder to hold the sub-key value.
                    oj_key = ObjectRecorder(new_node, {}, sub_key)
                    return stitch_node_inner(new_node, {"self": val, "key": oj_key}, sg)

            else:
                # This is the case that the picked key is not found in the
                # dictionary. In this case, we just ignore the pick and don't
                # stitch any further.
                pass
        else:
            # In this case, we don't have a constant key. This case comes up
            # very often as it is used for run colors. in particular, run colors
            # in the table is something like {"run_name":
            # "run_color"}.pick([long_graph].name()) In this case, we don't know
            # which key to pick, so we just ignore. In the future, we may need
            # to act as if we are picking every key?
            pass
    elif node.from_op.name.endswith("map"):
        fn = inputs[1].val
        if fn is None:
            raise errors.WeaveInternalError("non-const not yet supported")
        # Return the resulting object from map!
        return subgraph_stitch(fn, {"row": inputs[0]}, sg)
    elif node.from_op.name.endswith("sort") or node.from_op.name.endswith("filter"):
        fn = inputs[1].val
        if fn is None:
            raise errors.WeaveInternalError("non-const not yet supported")
        subgraph_stitch(fn, {"row": inputs[0]}, sg)
        # # Return the original object
        return inputs[0]
    elif node.from_op.name.endswith("groupby"):
        fn = inputs[1].val
        if fn is None:
            raise errors.WeaveInternalError("non-const not yet supported")
        # The output of the subgraph function is the group key which becomes
        # the groupKey tag.
        groupkey = subgraph_stitch(fn, {"row": inputs[0]}, sg)
        inputs[0].tags["groupKey"] = groupkey
        # And we return the original object
        return inputs[0]
    elif node.from_op.name.endswith("join"):
        fn1 = inputs[2].val
        fn2 = inputs[3].val
        if fn1 is None:
            raise errors.WeaveInternalError("non-const not yet supported")
        if fn2 is None:
            raise errors.WeaveInternalError("non-const not yet supported")
        alias1 = inputs[4].val
        alias2 = inputs[5].val
        if alias1 is None:
            raise errors.WeaveInternalError("non-const not yet supported")
        if alias2 is None:
            raise errors.WeaveInternalError("non-const not yet supported")
        joinKey1 = subgraph_stitch(fn1, {"row": inputs[0]}, sg)
        joinKey2 = subgraph_stitch(fn2, {"row": inputs[1]}, sg)
        result = LiteralDictObjectRecorder(
            node, val={alias1: inputs[0], alias2: inputs[1]}
        )
        # We only pass joinKey1 in as tag. I think this is ok, as stitch has already determined
        # that everything in the tag is a necessary column (we needed those columns to produce
        # the join key in the first place).
        # But I'm not totally sure!
        result.tags["joinObj"] = joinKey1
        return result
    elif node.from_op.name.endswith("joinAll"):
        fn = inputs[1].val
        if fn is None:
            raise errors.WeaveInternalError("non-const not yet supported")
        # The output of the subgraph function is the joinKey which becomes
        # the joinkey tag.
        # TODO: Do we need to do this for each input?
        joinKey = subgraph_stitch(fn, {"row": inputs[0]}, sg)
        inputs[0].tags["joinObj"] = joinKey
        # And we return the original object
        return inputs[0]
    elif node.from_op.name == "execute":
        # Special case where the execute op is used to execute a subgraph.
        # We want the results to flow through
        fn_node = inputs[0].val
        if fn_node is None:
            raise errors.WeaveInternalError("execute function should not be none")
        return subgraph_stitch(fn_node, {}, sg)
    elif len(inputs) == 0:
        # op does not have any inputs, just track its downstream calls
        return ObjectRecorder(node)
    # Otherwise, not a special op, track its call.
    return inputs[0].call_node(node, input_dict)


def _apply_tag_rules_to_stitch_result(
    result: ObjectRecorder,
    op: op_def.OpDef,
    inputs: list[ObjectRecorder],
    input_names: list[str],
) -> None:
    # Tag logic
    # If the op is a mapped, derived op, then we need the tags to flow
    # internally. We know we need to do this because there is special tag
    # handling logic in the mapped ops which does a parallel job. Note: This is
    # probably something that needs to be done for arrow as well.
    should_tag_with_inputs = False
    should_flow_tags = False
    if op.derived_from and op.derived_from.derived_ops.get("mapped"):
        if opdef_util.should_tag_op_def_outputs(op.derived_from):
            should_tag_with_inputs = True
        elif opdef_util.should_flow_tags(op.derived_from):
            should_flow_tags = True
    # Always do this, even for mapped
    if opdef_util.should_tag_op_def_outputs(op):
        should_tag_with_inputs = True
    elif opdef_util.should_flow_tags(op):
        should_flow_tags = True

    if should_tag_with_inputs:
        result.tags.update(inputs[0].tags)
        result.tags[input_names[0]] = inputs[0]
    elif should_flow_tags:
        result.tags.update(inputs[0].tags)


def stitch_node(
    node: graph.OutputNode, input_dict: dict[str, ObjectRecorder], sg: StitchedGraph
) -> ObjectRecorder:
    op = registry_mem.memory_registry.get_op(node.from_op.name)
    input_names = list(input_dict.keys())
    inputs = list(input_dict.values())

    result = stitch_node_inner(node, input_dict, sg)
    _apply_tag_rules_to_stitch_result(result, op, inputs, input_names)

    return result



from pydantic import BaseModel, create_model

from weave.legacy.weave import weave_types as types
from . import infer_types


def weave_type_to_pydantic(
    property_types: dict[str, types.Type], name: str
) -> BaseModel:
    field_types = {}
    for k, v in property_types.items():
        if isinstance(v, types.TypedDict):
            field_types[k] = weave_type_to_pydantic(v.property_types, "TypedDict")
        else:
            instance_classes = v._instance_classes()
            if not instance_classes:
                raise ValueError(f"Cannot convert {v} to pydantic")
            field_types[k] = instance_classes[0]  # type: ignore
    field_defs = {k: (v, ...) for k, v in field_types.items()}
    return create_model(name, **field_defs)  # type: ignore


def pydantic_class_to_attr_types(
    pydantic_class: type[BaseModel],
) -> dict[str, types.Type]:
    attr_types = {}
    for field_name, field in pydantic_class.model_fields.items():
        if field.annotation:
            attr_types[field_name] = infer_types.python_type_to_type(field.annotation)
    return attr_types



import time

import numpy as np
import wandb

import weave
from weave.legacy.weave.language_features.tagging import make_tag_getter_op
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType
from weave.legacy.weave.ops_arrow.list_ops import filter
from weave.legacy.weave.ops_domain import wbmedia
from weave.legacy.weave.ops_domain.wandb_domain_gql import _make_alias
from weave.legacy.weave.weave_internal import make_const_node


def use_static_artifact_node(
    fake_wandb,
    entity_name="test_entity",
    project_name="test_project",
    collection_name="joined_table_artifact",
    version="latest",
) -> weave.legacy.weave.graph.Node:
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_5702147f0293fd7538d402af13069708": {
                "id": "p1",
                "name": project_name,
                "entity": {"id": "e1", "name": entity_name},
                _make_alias(
                    f'name: "{collection_name}:{version}"', prefix="artifact"
                ): {
                    "id": "a1",
                    "versionIndex": "0",
                    "commitHash": "1478438ajsdfjj3kj1nn1nj",
                    "artifactSequence": {
                        "id": "c1",
                        "name": collection_name,
                        "project": {
                            "id": "p1",
                            "name": project_name,
                            "entity": {"id": "e1", "name": entity_name},
                        },
                        "defaultArtifactType": {
                            "id": "at1",
                            "name": "art_type",
                            "project": {
                                "id": "p1",
                                "name": project_name,
                                "entity": {"id": "e1", "name": entity_name},
                            },
                        },
                    },
                },
            }
        }
    )
    return weave.legacy.weave.ops.project(entity_name, project_name).artifactVersion(
        collection_name, version
    )


def test_wb_joined_table(fake_wandb):
    art_node = use_static_artifact_node(
        fake_wandb, collection_name="joined_table_artifact"
    )
    rows_node = art_node.file("table.joined-table.json").joinedTable().rows()
    cell_node = rows_node[0]
    assert weave.use(rows_node.count()) == 15
    assert weave.use(cell_node) == {
        "0": {"t_1_to_t_2_card": "1-1", "val": "a2"},
        "1": {"t_1_to_t_2_card": "1-1", "val": "a1"},
    }


def test_wb_partitioned_table(fake_wandb):
    art_node = use_static_artifact_node(
        fake_wandb, collection_name="partitioned_table_artifact"
    )
    rows_node = art_node.file("table.partitioned-table.json").partitionedTable().rows()
    cell_node = rows_node[0]
    assert weave.use(rows_node.count()) == 3
    assert weave.use(cell_node) == {"a": 1.0, "b": 2.0, "c": 3.0}


def test_convert_optional_list_cell(fake_wandb):
    tab = wandb.Table(columns=["a"])
    tab.add_data([wandb.Html("<p>hello</p>")])
    art = wandb.Artifact("test_name", "test_type")
    art.add(tab, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)
    file_node = art_node.file("table.table.json")
    table_node = file_node.table()
    table_rows = table_node.rows()
    awl = weave.use(table_rows)
    # The wandb library makes everything optional, which is what we want to
    # test here.
    # Make sure the mapped Weave1 type is what we expect.
    assert awl.object_type == weave.types.TypedDict(
        {
            "a": weave.types.optional(
                weave.types.List(wbmedia.HtmlArtifactFileRef.WeaveType())
            )
        }
    )
    assert awl.to_pylist_raw() == [
        {
            "a": [
                {
                    "artifact": f"wandb-artifact:///test_entity/test_project/test_name:{art.commit_hash}",
                    "path": "media/html/03ac15e611be692f058e.html",
                    "sha256": "d4935b7d4e8f30952d5869122ca6793114936be8bf156dd936b6794fb6715e02",
                }
            ]
        }
    ]


def _quick_image(val: int):
    return wandb.Image(np.ones((32, 32)) * val)


def test_join_table_with_images(fake_wandb):
    table1 = wandb.Table(
        columns=["name", "image"], data=[["a", _quick_image(1)], ["b", _quick_image(2)]]
    )
    table2 = wandb.Table(columns=["name", "score"], data=[["a", 1], ["b", 2]])
    jt = wandb.JoinedTable(table1, table2, "name")

    art = wandb.Artifact("test_name", "test_type")
    art.add(jt, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)
    file_node = art_node.file("table.joined-table.json")
    table_node = file_node.table()
    table_rows = table_node.rows()
    awl = weave.use(table_rows)
    assert awl.object_type == TaggedValueType(
        weave.types.TypedDict(
            {
                "joinObj": weave.types.optional(weave.types.String()),
            }
        ),
        weave.types.TypedDict(
            {
                "0": weave.types.optional(
                    weave.types.TypedDict(
                        {
                            "name": weave.types.optional(weave.types.String()),
                            "image": weave.types.optional(
                                wbmedia.ImageArtifactFileRef.WeaveType()
                            ),
                        }
                    )
                ),
                "1": weave.types.optional(
                    weave.types.TypedDict(
                        {
                            "name": weave.types.optional(weave.types.String()),
                            "score": weave.types.optional(weave.types.Number()),
                        }
                    )
                ),
            }
        ),
    )
    assert awl.to_pylist_raw() == [
        {
            "_tag": {"joinObj": "a"},
            "_value": {
                "0": {
                    "name": "a",
                    "image": {
                        "artifact": f"wandb-artifact:///test_entity/test_project/test_name:{art.commit_hash}",
                        "path": "media/images/9d4f26b99a1d4d044b6c.png",
                        "format": "png",
                        "height": 32,
                        "width": 32,
                        "sha256": "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
                        "boxes": {},
                        "masks": {},
                    },
                },
                "1": {"name": "a", "score": 1.0},
            },
        },
        {
            "_tag": {"joinObj": "b"},
            "_value": {
                "0": {
                    "name": "b",
                    "image": {
                        "artifact": f"wandb-artifact:///test_entity/test_project/test_name:{art.commit_hash}",
                        "path": "media/images/7fd26b0af1228fa077bb.png",
                        "format": "png",
                        "height": 32,
                        "width": 32,
                        "sha256": "61cd2467cff9f0666c730c57d065cfe834765ba26514b46f91735c750676876a",
                        "boxes": {},
                        "masks": {},
                    },
                },
                "1": {"name": "b", "score": 2.0},
            },
        },
    ]


def test_join_table_with_numeric_columns(fake_wandb):
    table = wandb.Table(columns=[1, 2], data=[[1, 2]])
    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)
    file_node = art_node.file("table.table.json")
    table_node = file_node.table()
    table_rows = table_node.rows()
    awl = weave.use(table_rows)
    # We want to convert these columns to strings - it is intentional that the
    # input is numbers but the output is strings. This is because Typescript,
    # JSON, and Python all handle numeric keys slightly diffently, so we settle
    # on always using the string representation.
    assert awl.object_type == weave.types.TypedDict(
        {
            "1": weave.types.optional(weave.types.Number()),
            "2": weave.types.optional(weave.types.Number()),
        }
    )
    assert awl.to_pylist_raw() == [{"1": 1, "2": 2}]


def test_metric_table_join(fake_wandb):
    table = wandb.Table(
        columns=["id", "label", "score_1", "score_2", "score_3"],
        data=[
            [1, "cat", 1.1, 2.1, 3.1],
            [2, "dog", 1.2, 2.2, 3.2],
            [3, "cat", 1.3, 2.3, 3.3],
            [4, "dog", 1.4, 2.4, 3.4],
            [5, "mouse", 1.5, 2.5, 3.5],
        ],
    )
    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)
    file_node = art_node.file("table.table.json")
    table_node = file_node.table()
    table_rows = table_node.rows().createIndexCheckpointTag()
    grouped = table_rows.groupby(lambda row: weave.legacy.weave.ops.dict_(label=row["label"]))
    sorted = grouped.sort(
        lambda row: weave.legacy.weave.ops.make_list(a=row.groupkey()["label"]), ["asc"]
    )
    group_col_0 = sorted[0].groupkey()["label"]
    group_col_1 = sorted[1].groupkey()["label"]
    group_col_2 = sorted[2].groupkey()["label"]
    res = weave.use([group_col_0, group_col_1, group_col_2])
    assert res == ["cat", "dog", "mouse"]


def test_empty_table(fake_wandb):
    table = wandb.Table(
        columns=["id", "label", "score_1", "score_2", "score_3"],
        data=[],
    )
    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)
    file_node = art_node.file("table.table.json")
    table_node = file_node.table()
    table_rows = table_node.rows().createIndexCheckpointTag()
    filtered = filter(table_rows, lambda row: row["label"] == "cat")
    grouped = filtered.groupby(lambda row: weave.legacy.weave.ops.dict_(label=row["label"]))
    sorted = grouped.sort(
        lambda row: weave.legacy.weave.ops.make_list(a=row.groupkey()["label"]), ["asc"]
    )
    res = weave.use(sorted)
    assert res.to_pylist_raw() == []


def test_join_group_combo(fake_wandb):
    table_1 = wandb.Table(
        columns=["id", "label", "score"],
        data=[
            [1, "A", 1],
            [2, "A", 2],
            [3, "B", 3],
            [4, "B", 4],
            [5, "C", 5],
            [6, "C", 6],
        ],
    )
    art_1 = wandb.Artifact("test_name_1", "test_type_1")
    art_1.add(table_1, "table_1")

    table_2 = wandb.Table(
        columns=["id", "label", "score"],
        data=[
            [1, "A", 10],
            [2, "B", 20],
            [3, "B", 30],
            [4, "C", 40],
            [5, "C", 50],
            [6, "A", 60],
        ],
    )
    art_2 = wandb.Artifact("test_name_2", "test_type_2")
    art_2.add(table_2, "table_2")

    art_1_node = fake_wandb.mock_artifact_as_node(art_1)
    art_2_node = fake_wandb.mock_artifact_as_node(art_2)
    table_1_rows = art_1_node.file("table_1.table.json").table().rows()
    table_2_rows = art_2_node.file("table_2.table.json").table().rows()
    list_of_tables = weave.legacy.weave.ops.make_list(a=table_1_rows, b=table_2_rows).dropna()
    joined_tables = list_of_tables.joinAll(
        lambda row: weave.legacy.weave.ops.make_list(a=row["id"]), False
    )
    indexed = joined_tables.createIndexCheckpointTag()
    grouped = indexed.groupby(lambda row: weave.legacy.weave.ops.dict_(label=row["label"][0]))
    sorted = grouped.sort(
        lambda row: weave.legacy.weave.ops.make_list(label=row.groupkey()["label"]),
        ["asc"],
    )
    assert weave.use(sorted.count()) == 3
    assert weave.use(sorted[2].groupkey()["label"]) == "C"
    assert weave.use(sorted[1]["score"]).to_pylist_notags() == [
        [3.0, 30.0],
        [4.0, 40.0],
    ]

    join_obj = sorted[0].joinObj()[0]
    assert weave.use(join_obj) == [1.0]

    from weave.legacy.weave import context_state

    _loading_builtins_token = context_state.set_loading_built_ins()
    tag_getter_op = make_tag_getter_op.make_tag_getter_op(
        "_ct_fake_run", weave.types.String()
    )
    context_state.clear_loading_built_ins(_loading_builtins_token)
    run_names = sorted[0]["score"].mapEach(lambda row: tag_getter_op(row))
    use_run_names = weave.use(run_names)
    assert use_run_names.to_pylist_notags() == [
        ["test_run", "test_run"],
        ["test_run", "test_run"],
    ]


def test_group_by_const(fake_wandb):
    columns = ["id", "label", "score"]
    data = [
        [1, "A", 1],
        [2, "A", 2],
        [3, "B", 3],
        [4, "B", 4],
        [5, "C", 5],
        [6, "C", 6],
    ]
    table_1 = wandb.Table(
        columns=columns,
        data=data,
    )
    art_1 = wandb.Artifact("test_name_1", "test_type_1")
    art_1.add(table_1, "table_1")
    art_1_node = fake_wandb.mock_artifact_as_node(art_1)
    table_1_rows = art_1_node.file("table_1.table.json").table().rows()
    grouped = table_1_rows.groupby(
        lambda row: weave.legacy.weave.ops.dict_(
            a=make_const_node(weave.types.Boolean(), True)
        )
    )
    assert weave.use(grouped).to_pylist_notags() == [
        [dict(zip(columns, row)) for row in data]
    ]


def test_symbols_in_name(fake_wandb):
    columns = ["id", "label", "score"]
    data = [
        [1, "A", 1],
        [2, "A", 2],
        [3, "B", 3],
        [4, "B", 4],
        [5, "C", 5],
        [6, "C", 6],
    ]
    table_1 = wandb.Table(
        columns=columns,
        data=data,
    )
    art_1 = wandb.Artifact("test_name_1", "test_type_1")
    name_with_all_symbols_and_spaces = "table_1!@#$%^&*( )_+`~[]{}|;':\",./<>?"
    art_1.add(table_1, name_with_all_symbols_and_spaces)
    art_1_node = fake_wandb.mock_artifact_as_node(art_1)
    table_1_rows = (
        art_1_node.file(f"{name_with_all_symbols_and_spaces}.table.json").table().rows()
    )
    assert weave.use(table_1_rows).to_pylist_notags() == [
        dict(zip(columns, row)) for row in data
    ]


def test_column_sort(fake_wandb):
    columns = ["name", "number", "float", "timestamp"]
    row1 = ["a", 0, 0.124122143123, time.time()]
    row2 = ["b", 1, 100.124123412341, time.time() + 10]
    row3 = ["c", 2, 10000.123412341234, time.time() + 20]
    data = [row1, row2, row3]

    table_1 = wandb.Table(
        columns=columns,
        data=data,
    )
    art_1 = wandb.Artifact("test_name_1", "test_type_1")
    art_1.add(table_1, "table_1")
    art_1_node = fake_wandb.mock_artifact_as_node(art_1)
    rows = art_1_node.file("table_1.table.json").table().rows()

    for col in columns:
        sorted = rows.sort(
            lambda row: weave.legacy.weave.ops.make_list(label=row[col]), ["desc"]
        )
        assert weave.use(sorted).to_pylist_notags() == [
            dict(zip(columns, row)) for row in [row3, row2, row1]
        ]

        sorted = sorted.sort(
            lambda row: weave.legacy.weave.ops.make_list(label=row[col]), ["asc"]
        )
        assert weave.use(sorted).to_pylist_notags() == [
            dict(zip(columns, row)) for row in data
        ]

    # Additional test sorting typed timestamps
    sorted = rows.sort(
        lambda row: weave.legacy.weave.ops.make_list(label=row["timestamp"].toTimestamp()),
        ["desc"],
    )
    assert weave.use(sorted).to_pylist_notags() == [
        dict(zip(columns, row)) for row in [row3, row2, row1]
    ]

    sorted = sorted.sort(
        lambda row: weave.legacy.weave.ops.make_list(label=row["timestamp"].toTimestamp()),
        ["asc"],
    )
    assert weave.use(sorted).to_pylist_notags() == [
        dict(zip(columns, row)) for row in data
    ]


def test_group_avg_sort_combo(fake_wandb):
    columns = ["id", "label", "score"]
    data = [
        [1, "A", 1],
        [2, "A", 2],
        [3, "B", 3],
        [4, "B", 4],
        [5, "C", 5],
        [6, "C", 6],
    ]

    table_1 = wandb.Table(
        columns=columns,
        data=data,
    )
    art_1 = wandb.Artifact("test_name_1", "test_type_1")
    art_1.add(table_1, "table_1")
    art_1_node = fake_wandb.mock_artifact_as_node(art_1)
    rows = art_1_node.file("table_1.table.json").table().rows()

    grouped = rows.groupby(lambda row: weave.legacy.weave.ops.dict_(label=row["label"]))
    sorted = grouped.sort(
        lambda row: weave.legacy.weave.ops.make_list(label=row["score"].avg()), ["asc"]
    )
    assert weave.use(sorted[2].groupkey()["label"]) == "C"



import json

import graphql
import wandb

import weave
from weave.legacy.weave import ops
from weave.legacy.weave.language_features.tagging import tagged_value_type
from weave.legacy.weave.ops_domain import wb_domain_types
from weave.legacy.weave.ops_primitives import _dict_utils

from ...legacy.weave import registry_mem
from ...tests import fixture_fakewandb as fwb

"""
Tests in this file whould be used to test the graphs that can be constructed
with the ops in `ops_domain`. This file compliments `test_wb.py`. `test_wb.py`
is inteded to test common WB app graphs that come from high traffic page (eg.
workspaces, artifact browser, etc.) and primarily tests for correctness of the
result. This file is intended to help author unit tests for all the different
ops as well as test for the correctness of the GQL query that is generated as a
result of executing such graph. One example is probivided below
`test_root_project_runs`, but many more should be added.
"""


def assert_gql_str_equal(gql_doc, exp_str):
    stripped_exp = graphql.language.print_ast(graphql.language.parse(exp_str))
    stripped_got = graphql.language.print_ast(
        graphql.language.parse(gql_doc.loc.source.body)
    )
    assert stripped_got == stripped_exp


def test_root_project_runs(fake_wandb):
    runs_node = ops.project("stacey", "mendeleev").runs()
    runs_0_id = runs_node[0].id()
    runs_1_iid = runs_node[1].internalId()
    runs_2_name = runs_node[2].name()
    concat_node = runs_0_id.add(runs_1_iid).add(runs_2_name)
    fake_wandb.fake_api.add_mock(
        lambda query, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                "id": 1,
                "name": "mendeleev",
                "entity": {
                    "id": 1,
                    "name": "stacey",
                },
                "runs_c1233b7003317090ab5e2a75db4ad965": {
                    "edges": [
                        {
                            "node": {
                                "id": "id_0",
                                "name": "2ed5xwpn",
                                "displayName": "crazy-cat-5",
                            }
                        },
                        {
                            "node": {
                                "id": "id_1",
                                "name": "1x36vcvi",
                                "displayName": "large-cat-6",
                            }
                        },
                        {
                            "node": {
                                "id": "id_2",
                                "name": "f54e6rg7",
                                "displayName": "small-cat-7",
                            }
                        },
                    ]
                },
            }
        }
    )
    assert weave.use(concat_node) == "2ed5xwpnid_1small-cat-7"
    log = fake_wandb.fake_api.execute_log()
    assert len(log) == 1
    assert_gql_str_equal(
        log[0]["gql"],
        # Note: the inner project/entity query is because it is part of the required fragment for runs
        # this could in theory change in the future.
        """
        query WeavePythonCG {
            project_518fa79465d8ffaeb91015dce87e092f: project(name: "mendeleev", entityName: "stacey") {
                id
                name
                runs_c1233b7003317090ab5e2a75db4ad965: runs(first: 100) {
                    edges {
                        node {
                            id
                            name
                            displayName
                        }
                    }
                }
            }
        }""",
    )


def test_root_project_concat(fake_wandb):
    runs_node_1 = ops.project("stacey", "mendeleev").filteredRuns("{}", "-createdAt")
    runs_node_2 = ops.project("stacey", "mendeleev").filteredRuns("{}", "-createdAt")
    fake_wandb.fake_api.add_mock(
        lambda query, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,
                "runs_261949318143369aa6c158af92afee03": {
                    "edges": [
                        {
                            "node": {
                                **fwb.run_payload,
                                "summaryMetricsSubset": '{"loss": 0.1}',
                                "summaryMetrics": '{"loss": 0.1}',
                            },
                        }
                    ]
                },
            }
        }
    )
    summary = (
        ops.make_list(a=runs_node_1, b=runs_node_2).concat().limit(50).summary()["loss"]
    )
    log = fake_wandb.fake_api.execute_log()
    assert len(log) == 1
    # The first request is to refineSummary, so we request summaryMetrics
    assert_gql_str_equal(
        log[0]["gql"],
        # Note: the inner project/entity query is because it is part of the required fragment for runs
        # this could in theory change in the future.
        """query WeavePythonCG {
            project_518fa79465d8ffaeb91015dce87e092f: project(name: "mendeleev", entityName: "stacey") {
                id
                name
                runs_261949318143369aa6c158af92afee03: runs(
                first: 100
                filters: "{}"
                order: "-createdAt"
                ) {
                edges {
                    node {
                    id
                    name
                    summaryMetrics
                    }
                }
                }
            }
            }""",
    )
    assert tagged_value_type.TaggedValueType(
        weave.types.TypedDict(property_types={"project": wb_domain_types.ProjectType}),
        weave.types.List(
            object_type=tagged_value_type.TaggedValueType(
                weave.types.TypedDict(
                    property_types={
                        "run": tagged_value_type.TaggedValueType(
                            weave.types.TypedDict(
                                property_types={"project": wb_domain_types.ProjectType}
                            ),
                            wb_domain_types.RunType,
                        )
                    }
                ),
                weave.types.Float(),
            )
        ),
    ).assign_type(summary.type)

    assert weave.use(summary) == [0.1, 0.1]
    # The second request is the graph we constructed above. Projection
    # pushdown should occur (we should request summaryMetrics with the keys
    # arg)
    assert len(log) == 2
    assert_gql_str_equal(
        log[1]["gql"],
        # Note: the inner project/entity query is because it is part of the required fragment for runs
        # this could in theory change in the future.
        """query WeavePythonCG {
            project_518fa79465d8ffaeb91015dce87e092f: project(name: "mendeleev", entityName: "stacey") {
                id
                name
                runs_261949318143369aa6c158af92afee03: runs(
                first: 100
                filters: "{}"
                order: "-createdAt"
                ) {
                edges {
                    node {
                    id
                    name
                    summaryMetricsSubset: summaryMetrics(keys: ["loss"])
                    project {
                        id
                        name
                        entity {
                        id
                        name
                        }
                    }
                    }
                }
                }
            }
            }""",
    )


def test_all_projects(fake_wandb):
    all_projects = ops.project_ops.root_all_projects().runs().flatten().name()
    fake_wandb.fake_api.add_mock(
        lambda query, ndx: {
            "instance": {
                "projects_500": {
                    "edges": [
                        {
                            "node": {
                                **fwb.project_payload,
                                "runs_c1233b7003317090ab5e2a75db4ad965": {
                                    "edges": [
                                        {
                                            "node": {
                                                **fwb.run_payload,
                                                "displayName": "crazy-cat-5",
                                            },
                                        }
                                    ]
                                },
                            }
                        }
                    ]
                }
            }
        }
    )
    assert weave.use(all_projects) == ["crazy-cat-5"]
    log = fake_wandb.fake_api.execute_log()
    assert len(log) == 1
    assert_gql_str_equal(
        log[0]["gql"],
        # Note: the inner project/entity query is because it is part of the required fragment for runs
        # this could in theory change in the future.
        """query WeavePythonCG {
            instance {
                projects_500: projects {
                    edges {
                        node {
                            id
                            name
                            runs_c1233b7003317090ab5e2a75db4ad965: runs(first: 100) {
                                edges {
                                    node {
                                        id
                                        name
                                        displayName
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }""",
    )


def test_rpt_op(fake_wandb):
    rpt_op = registry_mem.memory_registry.get_op("rpt_weekly_users_by_country_by_repo")
    data = rpt_op("stacey")["rows"]["user_fraction"]
    fake_wandb.fake_api.add_mock(
        lambda query, ndx: {
            "repoInsightsPlotData_541e3882f7cccacef0f697358bac12e3": {
                "edges": [
                    {"node": {"row": [0.5, "US", 1674068711.643377, "pytorch"]}},
                    {"node": {"row": [0.75, "CA", 1674068711.643377, "pytorch"]}},
                ],
                "schema": [
                    {"Name": "user_fraction", "Type": "FLOAT"},
                    {"Name": "country", "Type": "STRING"},
                    {"Name": "created_week", "Type": "TIMESTAMP"},
                    {"Name": "framework", "Type": "STRING"},
                ],
                "isNormalizedUserCount": True,
            }  #
        }
    )
    assert weave.use(data) == [0.5, 0.75]
    log = fake_wandb.fake_api.execute_log()
    assert len(log) == 1
    assert_gql_str_equal(
        log[0]["gql"],
        # Note: the inner project/entity query is because it is part of the required fragment for runs
        # this could in theory change in the future.
        """query WeavePythonCG {
            repoInsightsPlotData_541e3882f7cccacef0f697358bac12e3: repoInsightsPlotData(
                plotName: "weekly_users_by_country_by_repo"
                repoName: "stacey"
                first: 100000
            ) {
                edges {
                    node {
                        row
                    }
                }
                schema
                isNormalizedUserCount
            }
        }""",
    )


def test_multi_root_merging(fake_wandb, cache_mode_minimal):
    fake_wandb.fake_api.add_mock(
        lambda query, ndx: {
            "project_8d1592567720841659de23c02c97d594": {
                **fwb.project_payload,
                "name": "p_0",
                "createdAt": "2020-01-01T00:00:00.000+00:00",
            },
            "project_3c237e5b25fed9a705b21513dd7921c6": {
                **fwb.project_payload,
                "name": "p_1",
                "runs_c1233b7003317090ab5e2a75db4ad965": {
                    "edges": [
                        {
                            "node": {
                                **fwb.run_payload,
                            }
                        }
                    ]
                },
            },
            "instance": {
                "projects_500": {
                    "edges": [
                        {
                            "node": {
                                **fwb.project_payload,
                                "name": "p_0",
                                "createdAt": "2021-01-01T00:00:00.000+00:00",
                            }
                        },
                        {
                            "node": {
                                **fwb.project_payload,
                                "name": "p_1",
                                "createdAt": "2022-01-01T00:00:00.000+00:00",
                            }
                        },
                    ]
                }
            },
        }
    )
    p_0_a = ops.project("e_0", "p_0")
    p_0_b = ops.project("e_0", "p_0")
    p_1_a = ops.project("e_1", "p_1")

    p_0_name = p_0_a.name()
    p_0_ca = p_0_b.createdAt().toNumber()
    p_1_a_runs = p_1_a.runs().count()

    r_p_a = ops.project_ops.root_all_projects()
    r_p_b = ops.project_ops.root_all_projects()
    r_p_a_name = r_p_a.name()
    r_p_b_ca = r_p_b.createdAt().toNumber()

    assert weave.use([p_0_name, p_0_ca, p_1_a_runs, r_p_a_name, r_p_b_ca]) == [
        "p_0",
        1577836800000,
        1,
        ["p_0", "p_1"],
        [1609459200000, 1640995200000],
    ]

    log = fake_wandb.fake_api.execute_log()
    assert len(log) == 1
    assert_gql_str_equal(
        log[0]["gql"],
        """
        query WeavePythonCG{
            project_8d1592567720841659de23c02c97d594:project(name:"p_0" entityName:"e_0"){
                id
                name
                createdAt
            }
            project_3c237e5b25fed9a705b21513dd7921c6:project(name:"p_1" entityName:"e_1"){
                id
                name
                runs_c1233b7003317090ab5e2a75db4ad965:runs(first:100){
                    edges{
                        node{
                            id
                            name
                        }
                    }
                }
            }
            instance{
                projects_500:projects {
                    edges{
                        node{
                            id
                            name
                            createdAt
                        }
                    }
                }
            }
        }
        """,
    )


def test_two_level_summary(fake_wandb):
    def _mocked_query(query, ndx):
        run_selections = (
            query["gql"]
            .definitions[0]
            .selection_set.selections[0]
            .selection_set.selections[2]
            .selection_set.selections[0]
            .selection_set.selections[0]
            .selection_set.selections
        )
        assert (
            len(
                list(
                    f
                    for f in run_selections
                    if f.name.value == "summaryMetrics"
                    and f.alias
                    and f.alias.value == "summaryMetricsSubset"
                )
            )
            == 1
        )
        return {
            "project_8d1592567720841659de23c02c97d594": {
                **fwb.project_payload,
                "name": "p_1",
                "runs_c1233b7003317090ab5e2a75db4ad965": {
                    "edges": [
                        {
                            "node": {
                                "name": "r_1",
                                "project": {"name": "p_1", "entity": {"name": "e_1"}},
                                "summaryMetricsSubset": json.dumps({"a": 1, "b": "x"}),
                            }
                        },
                        {
                            "node": {
                                "name": "r_1",
                                "project": {"name": "p_1", "entity": {"name": "e_1"}},
                                "summaryMetricsSubset": json.dumps({"a": 2, "b": "y"}),
                            }
                        },
                    ]
                },
            }
        }

    fake_wandb.fake_api.add_mock(_mocked_query)
    n = (
        ops.project("e_0", "p_0")
        .runs()
        .filter(lambda r: r.summary()["a"] == 1)
        .summary()["b"]
    )
    assert weave.use(n) == ["x"]


def test_escaped_gql_query(fake_wandb):
    response = {
        "project_8d1592567720841659de23c02c97d594": {
            "id": "UHJvamVjdDp2MTpzYWdlbWFrZXItcGVvcGxlLXZlaGljbGUtY2xhc3Mtc3BsaXR0aW5nOmFjdHVhdGVhaQ==",
            "name": "",
            "entity": {"id": "RW50aXR5OjE0NzUxNw==", "name": "e_0"},
            "runs_261949318143369aa6c158af92afee03": {"edges": []},
        }
    }

    fake_wandb.fake_api.add_mock(lambda q, ix: response)

    key = "Tables/NMS_0\\.45_IOU_0\\.5"
    node = (
        ops.project("e_0", "p_0")
        .filteredRuns("{}", "-createdAt")
        .limit(1)
        .summary()[key]
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()
        .index(6)["label"]
    )

    log = fake_wandb.fake_api.execute_log()
    assert_gql_str_equal(
        log[1]["gql"],
        # Note: the inner project/entity query is because it is part of the required fragment for runs
        # this could in theory change in the future.
        """
        query WeavePythonCG {
        project_8d1592567720841659de23c02c97d594: project(
            name: "p_0"
            entityName: "e_0"
        ) {
            id
            name
            runs_261949318143369aa6c158af92afee03: runs(
            first: 100
            filters: "{}"
            order: "-createdAt"
            ) {
            edges {
                node {
                id
                name
                summaryMetricsSubset: summaryMetrics(
                    keys: ["Tables/NMS_0.45_IOU_0.5"]
                )
                project {
                    id
                    name
                    entity {
                    id
                    name
                    }
                }
                }
            }
            }
        }
        }
        """,
    )


def test_null_propagation_through_nonnull_gql_ops(fake_wandb):
    fake_wandb.fake_api.add_mock(
        lambda q, ix: {"project_8d1592567720841659de23c02c97d594": None}
    )

    # this should fail?
    node = ops.project("e_0", "p_0").run("r_0").name()
    assert weave.use(node) == None



import numpy as np
import pytest
from PIL import Image

from weave.legacy.weave import storage
from weave.legacy.weave.ops_primitives import api as weave


def test_image():
    im = Image.linear_gradient("L").rotate(4)
    ref = storage.save(im)
    im2 = storage.get(ref)
    assert (np.array(im) == np.array(im2)).all()


def test_list_image_same_sizes():
    ims = [Image.linear_gradient("L").rotate(i * 4) for i in range(3)]
    ref = storage.save(ims)
    ims2 = storage.get(str(ref))
    assert len(ims) == len(ims2)
    for im, im2 in zip(ims, ims2):
        assert (np.array(im) == np.array(im2)).all()


@pytest.mark.skip("type merging is just the simple union algorithm for now")
def test_list_image_different_sizes_merges():
    ims = [
        Image.linear_gradient("L").resize((256 - i, 256)).rotate(i * 4)
        for i in range(3)
    ]
    # We'd need to change this line if we want to use this test
    assert weave.type_of(ims) == weave.types.Invalid()



import datetime
import os
import time

import numpy as np
import pytest
import wandb
from bokeh.plotting import figure
from wandb import Artifact
from wandb import data_types as wb_data_types
from wandb.sdk.artifacts.artifact_state import ArtifactState
from wandb.sdk.data_types._dtypes import TypeRegistry as SDKTypeRegistry

import weave
from weave.legacy.weave import artifact_fs
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.artifact_wandb import WandbArtifact, WeaveWBArtifactURI
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType
from weave.legacy.weave.ops_domain.wbmedia import ImageArtifactFileRefType
from weave.legacy.weave.ops_primitives import file
from weave.legacy.weave.wandb_client_api import wandb_gql_query
from weave.legacy.weave.wandb_util import weave0_type_json_to_weave1_type

from ...tests.fixture_fakewandb import FakeApi


class RandomClass:
    pass


def make_image():
    return wandb.Image(np.random.randint(0, 255, (32, 32)))


def make_audio():
    return wandb.Audio(np.random.uniform(-1, 1, 44100), 44100)


HTML_STRING = "<body><h1>Hello this is some HTML with symbols and emoji '$!$!@#&#$%^??%$# 🤠</h1></body>"


def make_html():
    return wandb.Html(HTML_STRING)


def make_bokeh():
    x = [1, 2, 3, 4, 5]
    y = [6, 7, 2, 4, 5]
    p = figure(title="simple line example", x_axis_label="x", y_axis_label="y")
    p.line(x, y, legend_label="Temp.", line_width=2)
    return wb_data_types.Bokeh(p)


def make_video(clean_up=True):
    with open("video.mp4", "w") as f:
        f.write("00000")
    vid = wandb.Video("video.mp4")
    if clean_up:
        os.remove("video.mp4")
    return vid


def make_object3d():
    return wandb.Object3D(
        np.array(
            [
                [0, 0, 0, 1],
                [0, 0, 1, 13],
                [0, 1, 0, 2],
                [0, 1, 0, 4],
            ]
        )
    )


def make_molecule(clean_up=True):
    with open("test_mol.pdb", "w") as f:
        f.write("00000")
    mol = wandb.Molecule("test_mol.pdb")
    if clean_up:
        os.remove("test_mol.pdb")
    return mol


@pytest.mark.parametrize(
    "sdk_obj, expected_type",
    [
        #
        # Primitive Types
        #
        (None, types.none_type),
        (True, types.Boolean()),
        (42, types.Number()),
        ("hello", types.String()),
        #
        # Container Types
        #
        ({"hello": "world"}, types.TypedDict({"hello": types.String()})),
        ([1, 2, 3], types.List(types.Number())),
        ([{"hello": "world"}], types.List(types.TypedDict({"hello": types.String()}))),
        #
        # Domain Types
        #
        (datetime.datetime.now(), types.Timestamp()),  # type: ignore
        # See comment in wandb_util.py and https://github.com/wandb/weave/pull/140
        (np.array([1, 2, 3]), types.List(types.UnknownType())),
        #
        # Media Types
        #
        (
            make_image(),
            weave.legacy.weave.ops.ImageArtifactFileRef.WeaveType(),  # type: ignore
        ),
        (
            make_audio(),
            weave.legacy.weave.ops.AudioArtifactFileRef.WeaveType(),  # type: ignore
        ),
        (
            make_html(),
            weave.legacy.weave.ops.HtmlArtifactFileRef.WeaveType(),  # type: ignore
        ),
        (
            make_bokeh(),
            weave.legacy.weave.ops.BokehArtifactFileRef.WeaveType(),  # type: ignore
        ),
        (
            make_video(),
            weave.legacy.weave.ops.VideoArtifactFileRef.WeaveType(),  # type: ignore
        ),
        (
            make_object3d(),
            weave.legacy.weave.ops.Object3DArtifactFileRef.WeaveType(),  # type: ignore
        ),
        (
            make_molecule(),
            weave.legacy.weave.ops.MoleculeArtifactFileRef.WeaveType(),  # type: ignore
        ),
        # See comment in wandb_util.py - this may change in the future
        # Temporarily disabled until we can figure out how to mock
        # types that need to reach into the artifact
        # (
        #     SDKClasses([{"id": 1, "name": "foo"}]),
        #     types.Number(),
        # ),
        #
        # Table Types
        # Leaving the table/key types out for now since there are not code paths
        # that exersize this. We will likely need to add these in the future,
        # but in the interest of incremental PRs, I'm leaving them out
        # TODO: 3 table Types: wandb_data_types._TableType, wandb_data_types._JoinedTableType, wandb_data_types._PartitionedTableType
        # TODO: 3 key Types:  wandb_data_types._PrimaryKeyType, wandb_data_types._ForeignKeyType, wandb_data_types._ForeignIndexType
        #
        # Legacy Fallback Types
        #
        (RandomClass(), types.UnknownType()),
    ],
)
def test_image(sdk_obj, expected_type, fake_wandb):
    art = wandb.Artifact("test", "test")
    obj_json = SDKTypeRegistry.type_of(sdk_obj).to_json(art)
    # Create an artifact that looks like it was loaded remotely so we can use it without mocking backend
    art._id = "1234567890"
    art._entity = "test"
    art._project = "test"
    art._name = "test:v0"
    art._version = "v0"
    art._state = ArtifactState.COMMITTED

    assert weave0_type_json_to_weave1_type(obj_json) == expected_type


def peak(v, m):
    return 1 - ((abs((2 * v) - m + 1) % m) / m)


def make_np_image(dim=128):
    return np.array(
        [
            [(peak(col, dim) * peak(row, dim)) ** 0.5 for col in range(dim)]
            for row in range(dim)
        ]
    )


def make_wb_image(use_middle=False, use_pixels=False):
    if use_pixels:
        position = (
            {
                "middle": [50, 50],
                "height": 10,
                "width": 20,
            }
            if use_middle
            else {
                "minX": 40,
                "maxX": 100,
                "minY": 30,
                "maxY": 50,
            }
        )
        domain = "pixel"
    else:
        position = (
            {
                "middle": [0.5, 0.5],
                "height": 0.5,
                "width": 0.25,
            }
            if use_middle
            else {
                "minX": 0.4,
                "maxX": 0.6,
                "minY": 0.3,
                "maxY": 0.7,
            }
        )
        domain = None
    return wandb.Image(
        make_np_image(),
        boxes={
            "box_set_1": {
                "box_data": [
                    {
                        "position": position,
                        "domain": domain,
                        "class_id": 0,
                        "scores": {"loss": 0.3, "gain": 0.7},
                        "box_caption": "a",
                    },
                ],
            },
            "box_set_2": {
                "box_data": [
                    {
                        "position": position,
                        "domain": domain,
                        "class_id": 2,
                        "scores": {"loss": 0.3, "gain": 0.7},
                    },
                ],
            },
        },
        masks={
            "mask_set_1": {
                "mask_data": np.array(
                    [[row % 4 for col in range(128)] for row in range(128)]
                )
            }
        },
        classes=[
            {"id": 0, "name": "c_zero"},
            {"id": 1, "name": "c_one"},
            {"id": 2, "name": "c_two"},
            {"id": 3, "name": "c_three"},
        ],
    )


def make_table():
    return wandb.Table(
        columns=["label", "image"],
        data=[
            ["a", make_wb_image(False, False)],
            ["b", make_wb_image(False, True)],
            ["c", make_wb_image(True, False)],
            ["d", make_wb_image(True, True)],
        ],
    )


def exp_raw_data(commit_hash: str):
    return [
        {
            "label": "a",
            "image": {
                "artifact": f"wandb-artifact:///test_entity/test_project/test_name:{commit_hash}",
                "path": "media/images/724389f96d933f4166db.png",
                "format": "png",
                "height": 128,
                "width": 128,
                "sha256": "82287185f849c094e7acf82835f0ceeb8a5d512329e8ce1da12e21df2e81e739",
                "boxes": {
                    "box_set_1": [
                        {
                            "box_caption": "a",
                            "class_id": 0,
                            "domain": None,
                            "position": {
                                "maxX": 0.6,
                                "maxY": 0.7,
                                "minX": 0.4,
                                "minY": 0.3,
                                "height": None,
                                "middle": None,
                                "width": None,
                            },
                            "scores": {"loss": 0.3, "gain": 0.7},
                        }
                    ],
                    "box_set_2": [
                        {
                            "box_caption": None,
                            "class_id": 2,
                            "domain": None,
                            "position": {
                                "maxX": 0.6,
                                "maxY": 0.7,
                                "minX": 0.4,
                                "minY": 0.3,
                                "height": None,
                                "middle": None,
                                "width": None,
                            },
                            "scores": {"loss": 0.3, "gain": 0.7},
                        }
                    ],
                },
                "masks": {
                    "mask_set_1": {
                        "_type": "mask",
                        "path": "media/images/mask/b98a2fc054512bf6450c.mask.png",
                        "sha256": "00c619b2faa45fdc9ce6de014e7aef7839c9de725bf78b528ef47d279039aacf",
                    }
                },
            },
        },
        {
            "label": "b",
            "image": {
                "artifact": f"wandb-artifact:///test_entity/test_project/test_name:{commit_hash}",
                "path": "media/images/724389f96d933f4166db.png",
                "format": "png",
                "height": 128,
                "width": 128,
                "sha256": "82287185f849c094e7acf82835f0ceeb8a5d512329e8ce1da12e21df2e81e739",
                "boxes": {
                    "box_set_1": [
                        {
                            "box_caption": "a",
                            "class_id": 0,
                            "domain": "pixel",
                            "position": {
                                "maxX": 100.0,
                                "maxY": 50.0,
                                "minX": 40.0,
                                "minY": 30.0,
                                "height": None,
                                "middle": None,
                                "width": None,
                            },
                            "scores": {"loss": 0.3, "gain": 0.7},
                        }
                    ],
                    "box_set_2": [
                        {
                            "box_caption": None,
                            "class_id": 2,
                            "domain": "pixel",
                            "position": {
                                "maxX": 100.0,
                                "maxY": 50.0,
                                "minX": 40.0,
                                "minY": 30.0,
                                "height": None,
                                "middle": None,
                                "width": None,
                            },
                            "scores": {"loss": 0.3, "gain": 0.7},
                        }
                    ],
                },
                "masks": {
                    "mask_set_1": {
                        "_type": "mask",
                        "path": "media/images/mask/b98a2fc054512bf6450c.mask.png",
                        "sha256": "00c619b2faa45fdc9ce6de014e7aef7839c9de725bf78b528ef47d279039aacf",
                    }
                },
            },
        },
        {
            "label": "c",
            "image": {
                "artifact": f"wandb-artifact:///test_entity/test_project/test_name:{commit_hash}",
                "path": "media/images/724389f96d933f4166db.png",
                "format": "png",
                "height": 128,
                "width": 128,
                "sha256": "82287185f849c094e7acf82835f0ceeb8a5d512329e8ce1da12e21df2e81e739",
                "boxes": {
                    "box_set_1": [
                        {
                            "box_caption": "a",
                            "class_id": 0,
                            "domain": None,
                            "position": {
                                "maxX": None,
                                "maxY": None,
                                "minX": None,
                                "minY": None,
                                "height": 0.5,
                                "middle": [0.5, 0.5],
                                "width": 0.25,
                            },
                            "scores": {"loss": 0.3, "gain": 0.7},
                        }
                    ],
                    "box_set_2": [
                        {
                            "box_caption": None,
                            "class_id": 2,
                            "domain": None,
                            "position": {
                                "maxX": None,
                                "maxY": None,
                                "minX": None,
                                "minY": None,
                                "height": 0.5,
                                "middle": [0.5, 0.5],
                                "width": 0.25,
                            },
                            "scores": {"loss": 0.3, "gain": 0.7},
                        }
                    ],
                },
                "masks": {
                    "mask_set_1": {
                        "_type": "mask",
                        "path": "media/images/mask/b98a2fc054512bf6450c.mask.png",
                        "sha256": "00c619b2faa45fdc9ce6de014e7aef7839c9de725bf78b528ef47d279039aacf",
                    }
                },
            },
        },
        {
            "label": "d",
            "image": {
                "artifact": f"wandb-artifact:///test_entity/test_project/test_name:{commit_hash}",
                "path": "media/images/724389f96d933f4166db.png",
                "format": "png",
                "height": 128,
                "width": 128,
                "sha256": "82287185f849c094e7acf82835f0ceeb8a5d512329e8ce1da12e21df2e81e739",
                "boxes": {
                    "box_set_1": [
                        {
                            "box_caption": "a",
                            "class_id": 0,
                            "domain": "pixel",
                            "position": {
                                "maxX": None,
                                "maxY": None,
                                "minX": None,
                                "minY": None,
                                "height": 10.0,
                                "middle": [50.0, 50.0],
                                "width": 20.0,
                            },
                            "scores": {"loss": 0.3, "gain": 0.7},
                        }
                    ],
                    "box_set_2": [
                        {
                            "box_caption": None,
                            "class_id": 2,
                            "domain": "pixel",
                            "position": {
                                "maxX": None,
                                "maxY": None,
                                "minX": None,
                                "minY": None,
                                "height": 10.0,
                                "middle": [50.0, 50.0],
                                "width": 20.0,
                            },
                            "scores": {"loss": 0.3, "gain": 0.7},
                        }
                    ],
                },
                "masks": {
                    "mask_set_1": {
                        "_type": "mask",
                        "path": "media/images/mask/b98a2fc054512bf6450c.mask.png",
                        "sha256": "00c619b2faa45fdc9ce6de014e7aef7839c9de725bf78b528ef47d279039aacf",
                    }
                },
            },
        },
    ]


def test_annotated_images_in_tables(fake_wandb):
    table = make_table()

    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)

    file_node = art_node.file("table.table.json")
    table_node = file_node.table()
    table_rows = table_node.rows()
    table_rows_type = table_node.rowsType()

    raw_data = weave.use(table_rows).to_pylist_notags()
    assert raw_data == exp_raw_data(art.commit_hash)

    ot = weave.use(table_rows_type).object_type
    assert ot == TaggedValueType(
        types.TypedDict(
            {"_ct_fake_run": types.String(), "_ct_fake_project": types.String()}
        ),
        weave.types.TypedDict(
            {
                "label": weave.types.UnionType(
                    weave.types.NoneType(), weave.types.String()
                ),
                "image": weave.types.UnionType(
                    ImageArtifactFileRefType(
                        boxLayers={"box_set_1": [0], "box_set_2": [2]},
                        boxScoreKeys=["loss", "gain"],
                        maskLayers={"mask_set_1": [0, 1, 2, 3]},
                        classMap={
                            "0": "c_zero",
                            "1": "c_one",
                            "2": "c_two",
                            "3": "c_three",
                        },
                    ),
                    weave.types.NoneType(),
                ),
            }
        ),
    )


def test_annotated_legacy_images_in_tables(fake_wandb):
    # Mocking this property makes the payload look like the legacy version.
    from wandb.sdk.data_types.image import _ImageFileType
    from wandb.sdk.data_types._dtypes import InvalidType

    def dummy_params(self):
        return {}

    def dummy_assign_type(self, wb_type=None):
        if isinstance(wb_type, _ImageFileType):
            return self
        return InvalidType()

    _ImageFileType.params = property(dummy_params)
    _ImageFileType.assign_type = dummy_assign_type

    table = make_table()

    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)

    file_node = art_node.file("table.table.json")
    table_node = file_node.table()
    table_rows = table_node.rows()
    table_rows_type = table_node.rowsType()

    raw_data = weave.use(table_rows).to_pylist_notags()
    assert raw_data == exp_raw_data(art.commit_hash)

    ot = weave.use(table_rows_type).object_type
    assert ot == TaggedValueType(
        types.TypedDict(
            {"_ct_fake_run": types.String(), "_ct_fake_project": types.String()}
        ),
        weave.types.TypedDict(
            {
                "label": weave.types.UnionType(
                    weave.types.NoneType(), weave.types.String()
                ),
                "image": weave.types.UnionType(
                    ImageArtifactFileRefType(
                        # Both box and mask layers have all the keys because
                        # we can't possible look at all the elements, so we assume
                        # they all may or may not have all the keys
                        boxLayers={
                            "box_set_1": [0, 1, 2, 3],
                            "box_set_2": [0, 1, 2, 3],
                        },
                        boxScoreKeys=["loss", "gain"],
                        maskLayers={"mask_set_1": [0, 1, 2, 3]},
                        classMap={
                            "0": "c_zero",
                            "2": "c_two",
                            "1": "c_one",
                            "3": "c_three",
                        },
                    ),
                    weave.types.NoneType(),
                ),
            }
        ),
    )


def make_simple_image_table():
    table = wandb.Table(columns=["label", "image"])
    for group in range(2):
        for item in range(3):
            table.add_data(
                f"{group}-{item}",
                wandb.Image(
                    np.ones((32, 32)) * group,
                    masks={
                        # Here, we add a mask set unique to each image to ensure that the mask is not considered in grouping
                        f"mask_set-{group}-{item}": {
                            "mask_data": np.array(
                                [[row % 4 for col in range(128)] for row in range(128)]
                            )
                        }
                    },
                ),
            )
    return table


def test_grouping_on_images(fake_wandb):
    table = make_simple_image_table()

    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)

    file_node = art_node.file("table.table.json")
    table_node = file_node.table()
    table_rows = table_node.rows()
    grouped = table_rows.groupby(
        lambda row: weave.legacy.weave.ops.dict_(g_image=row["image"])
    )

    raw_data = weave.use(grouped).to_pylist_notags()
    assert len(raw_data) == 2
    assert [[row["label"] for row in group] for group in raw_data] == [
        ["0-0", "0-1", "0-2"],
        ["1-0", "1-1", "1-2"],
    ]

    group_keys = weave.use(grouped.groupkey()).to_pylist_notags()
    assert [key["g_image"]["sha256"] for key in group_keys] == [
        "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
        "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    ]


exp_join2_labels = [
    ["0-0", "0-2"],
    ["0-0", "0-1"],
    ["0-0", "0-0"],
    ["0-1", "0-2"],
    ["0-1", "0-1"],
    ["0-1", "0-0"],
    ["0-2", "0-2"],
    ["0-2", "0-1"],
    ["0-2", "0-0"],
    ["1-0", "1-2"],
    ["1-0", "1-1"],
    ["1-0", "1-0"],
    ["1-1", "1-2"],
    ["1-1", "1-1"],
    ["1-1", "1-0"],
    ["1-2", "1-2"],
    ["1-2", "1-1"],
    ["1-2", "1-0"],
]

exp_join2_tag_shas = [
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
]

exp_joinall_labels = [
    ["0-0", "0-2"],
    ["0-0", "0-1"],
    ["0-0", "0-0"],
    ["0-1", "0-2"],
    ["0-1", "0-1"],
    ["0-1", "0-0"],
    ["0-2", "0-2"],
    ["0-2", "0-1"],
    ["0-2", "0-0"],
    ["1-0", "1-2"],
    ["1-0", "1-1"],
    ["1-0", "1-0"],
    ["1-1", "1-2"],
    ["1-1", "1-1"],
    ["1-1", "1-0"],
    ["1-2", "1-2"],
    ["1-2", "1-1"],
    ["1-2", "1-0"],
]

exp_joinall_tag_shas = [
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "1237db9e0c3d396728f7f4077f62b6283118c08fbfdced1e99e33205c270bd27",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
    "e7bdc527afd649f51950b4524b0c15aecaf7f484448a6cdfcdc2ecd9bba0f5a7",
]

exp_join_length = (3 * 3) * 2  # 3 copies of 2 unique images


def make_join_table_row_nodes(fake_wandb):
    table_1 = make_simple_image_table()
    table_2 = make_simple_image_table()

    art_1 = wandb.Artifact("test_name_1", "test_type_1")
    art_1.add(table_1, "table_1")

    art_2 = wandb.Artifact("test_name_2", "test_type_2")
    art_2.add(table_2, "table_2")

    art_1_node = fake_wandb.mock_artifact_as_node(art_1)
    art_2_node = fake_wandb.mock_artifact_as_node(art_2)

    file_1_node = art_1_node.file("table_1.table.json")
    table_1_node = file_1_node.table()
    table_1_rows = table_1_node.rows()

    file_2_node = art_2_node.file("table_2.table.json")
    table_2_node = file_2_node.table()
    table_2_rows = table_2_node.rows()

    return table_1_rows, table_2_rows


def test_join_all_on_images(fake_wandb):
    table_1_rows, table_2_rows = make_join_table_row_nodes(fake_wandb)

    rows = weave.legacy.weave.ops.make_list(a=table_1_rows, b=table_2_rows)

    joined = rows.joinAll(lambda row: row["image"], True)

    raw_data = weave.use(joined).to_pylist_notags()
    assert len(raw_data) == exp_join_length
    assert [row["label"] for row in raw_data] == exp_joinall_labels

    group_keys = weave.use(joined.joinObj()).to_pylist_notags()
    assert [key["sha256"] for key in group_keys] == exp_joinall_tag_shas


def test_join_2_on_images(fake_wandb):
    table_1_rows, table_2_rows = make_join_table_row_nodes(fake_wandb)

    joined = table_1_rows.join(
        table_2_rows,
        lambda row: row["image"],
        lambda row: row["image"],
        "t1",
        "t2",
        True,
        True,
    )

    raw_data = weave.use(joined).to_pylist_notags()
    assert len(raw_data) == exp_join_length
    got = [[row["t1"]["label"], row["t2"]["label"]] for row in raw_data]
    print("GOT", got)
    assert [
        [row["t1"]["label"], row["t2"]["label"]] for row in raw_data
    ] == exp_join2_labels
    group_keys = weave.use(joined.joinObj()).to_pylist_notags()
    got = [key["sha256"] for key in group_keys]
    print("GOT", got)
    assert [key["sha256"] for key in group_keys] == exp_join2_tag_shas


def test_html_encoding_decoding(fake_wandb):
    # Familiar wandb sdk code:
    html = make_html()

    art = wandb.Artifact("test_name", "test_type")
    art.add(html, "html")

    # new helper:
    art_node = fake_wandb.mock_artifact_as_node(art)

    # Weave Query off table:
    file_node = art_node.file("media/html/349dc3aced6290aac120.html")
    contents = file.file_contents(file_node)
    result = weave.use(contents)
    assert HTML_STRING in result


def test_media_logging_to_history(user_by_api_key_in_env, cache_mode_minimal):
    # TODO: Make this test exercise both the parquet and liveset
    # paths. Also test for values.
    log_dict = {
        "image": make_image(),  # make_wb_image(),
        "audio": make_audio(),
        "html": make_html(),
        "bokeh": make_bokeh(),
        "video": make_video(False),
        "object3d": make_object3d(),
        "molecule": make_molecule(False),
        "table": make_table(),
        "simple_image_table": make_simple_image_table(),
        "all_types_list": [
            make_image(),  # make_wb_image(),
            make_audio(),
            make_html(),
            make_bokeh(),
            make_video(False),
            make_object3d(),
            make_molecule(False),
            make_table(),
            make_simple_image_table(),
        ],
    }
    run = wandb.init(project="test_media_logging_to_history")
    run.log(log_dict)
    run.finish()

    # wait for history to be uploaded
    def wait_for():
        history = get_raw_gorilla_history(run.entity, run.project, run.id)
        return (
            len(history["parquetHistory"]["parquetUrls"]) > 0
            or len(history["parquetHistory"]["liveData"]) > 0
        )

    wait_for_x_times(wait_for)

    run_node = weave.legacy.weave.ops.project(run.entity, run.project).run(run.id)

    for history_op_name in ["history3", "history"]:
        history_node = run_node._get_op(history_op_name)()
        mapped_node = history_node.map(
            lambda row: weave.legacy.weave.ops.dict_(
                **{key: row[key] for key in log_dict.keys()}
            )
        )

        history = weave.use(mapped_node)
        if history_op_name == "history3":
            history = history.to_pylist_notags()

        assert len(history) == 1

        # Test image annotations
        # This will not work in W1 See comment in
        # run_history_v3_parquet_stream_optimized.py
        #
        # image_node = history_node["image"]
        # image_use = weave.use(image_node)
        # if history_op_name == "history3":
        #     image_use = image_use.to_pylist_notags()
        #     masks = image_use[0]["masks"]
        #     boxes = image_use[0]["boxes"]
        # else:
        #     masks = image_use[0].masks
        #     boxes = image_use[0].boxes
        # assert len(masks.keys()) > 0
        # assert len(boxes.keys()) > 0

        # Test file path access
        audio_node = history_node["audio"]
        audio_use = weave.use(audio_node)
        if history_op_name == "history3":
            audio_use = audio_use.to_pylist_notags()
            path = audio_use[0]["path"]
        else:
            path = audio_use[0].path
        file_node = audio_node[0].artifactVersion().file(path)
        file_use = weave.use(file_node)
        assert file_use != None

        # Table is a special case because it has a rowsType
        table_type_node = history_node["table"].table().rowsType()
        table_use = weave.use(table_type_node)
        assert table_use != None

    os.remove("video.mp4")
    os.remove("test_mol.pdb")


def get_raw_gorilla_history(entity_name, project_name, run_name):
    query = """query WeavePythonCG($entityName: String!, $projectName: String!, $runName: String! ) {
            project(name: $projectName, entityName: $entityName) {
                run(name: $runName) {
                    historyKeys
                    parquetHistory(liveKeys: ["_timestamp"]) {
                        liveData
                        parquetUrls
                    }
                }
            }
    }"""
    variables = {
        "entityName": entity_name,
        "projectName": project_name,
        "runName": run_name,
    }
    res = wandb_gql_query(query, variables)
    return res.get("project", {}).get("run", {})


def wait_for_x_times(for_fn, times=10, wait=1):
    done = False
    while times > 0 and not done:
        times -= 1
        done = for_fn()
        time.sleep(wait)
    assert times > 0



import random

import pytest

from weave.trace import refs
from weave.trace.weave_client import sanitize_object_name
from weave.trace_server import refs_internal

quote = refs_internal.extra_value_quoter


def test_isdescended_from():
    a = refs.ObjectRef(
        entity="e", project="p", name="n", digest="v", extra=["attr", "x2"]
    )
    b = refs.ObjectRef(
        entity="e",
        project="p",
        name="n",
        digest="v",
        extra=["attr", "x2", "attr", "x4"],
    )
    assert a.is_descended_from(b) == False
    assert b.is_descended_from(a) == True


def string_with_every_char(disallowed_chars=[]):
    char_codes = list(range(256))
    random.shuffle(char_codes)
    return "".join(chr(i) for i in char_codes if chr(i) not in disallowed_chars)


def test_ref_parsing_external_invalid():
    with pytest.raises(Exception):
        ref_start = refs.ObjectRef(
            entity="entity",
            project="project",
            name=string_with_every_char(),
            digest="1234567890",
            extra=("key", string_with_every_char()),
        )


def test_ref_parsing_external_sanitized():
    ref_start = refs.ObjectRef(
        entity="entity",
        project="project",
        name=sanitize_object_name(string_with_every_char()),
        digest="1234567890",
        extra=("key", string_with_every_char()),
    )

    ref_str = ref_start.uri()
    exp_ref = f"{refs_internal.WEAVE_SCHEME}:///{ref_start.entity}/{ref_start.project}/object/{ref_start.name}:{ref_start.digest}/{ref_start.extra[0]}/{quote(ref_start.extra[1])}"
    assert ref_str == exp_ref

    parsed = refs.parse_uri(ref_str)
    assert parsed == ref_start


def test_ref_parsing_internal_invalid():
    with pytest.raises(Exception):
        ref_start = refs_internal.InternalObjectRef(
            project_id="project",
            name=string_with_every_char(),
            version="1234567890",
            extra=("key", string_with_every_char()),
        )


def test_ref_parsing_internal_sanitized():
    ref_start = refs_internal.InternalObjectRef(
        project_id="project",
        name=sanitize_object_name(string_with_every_char()),
        version="1234567890",
        extra=["key", string_with_every_char()],
    )

    ref_str = ref_start.uri()
    exp_ref = f"{refs_internal.WEAVE_INTERNAL_SCHEME}:///{ref_start.project_id}/object/{ref_start.name}:{ref_start.version}/{ref_start.extra[0]}/{quote(ref_start.extra[1])}"
    assert ref_str == exp_ref

    parsed = refs_internal.parse_internal_uri(ref_str)
    assert parsed == ref_start



import weave
from weave.legacy.weave import ops_arrow


def test_cond_basic():
    assert weave.use(weave.legacy.weave.ops.cond({"a": True}, {"a": 5})) == 5
    assert weave.use(weave.legacy.weave.ops.cond({"a": False}, {"a": 5})) == None
    assert (
        weave.use(weave.legacy.weave.ops.cond({"a": False, "b": True}, {"a": 5, "b": 6})) == 6
    )


def test_cond_mapped_vector():
    conds = weave.save(
        ops_arrow.to_arrow(
            [{"a": True, "b": False}, {"a": False, "b": False}, {"a": False, "b": True}]
        ),
        "conds",
    )
    assert weave.use(conds.cond({"a": 5, "b": 6})).to_pylist_raw() == [5, None, 6]


def test_cond_mapped_vector_arr_value():
    conds = weave.save(
        ops_arrow.to_arrow(
            [{"a": True, "b": False}, {"a": False, "b": False}, {"a": False, "b": True}]
        ),
        "conds",
    )
    assert weave.use(conds.cond({"a": [1, 2], "b": [5, 6]})).to_pylist_raw() == [
        [1, 2],
        None,
        [5, 6],
    ]


def test_cond_vector():
    conds = weave.save(
        ops_arrow.to_arrow(
            [
                {"a": True, "b": False, "val_a": 5, "val_b": 6},
                {"a": False, "b": False, "val_a": 7, "val_b": 8},
                {"a": False, "b": True, "val_a": 9, "val_b": 10},
            ]
        ),
        "conds",
    )
    assert weave.use(
        conds.map(
            lambda row: weave.legacy.weave.ops.cond(
                weave.legacy.weave.ops.dict_(**{"a": row["a"], "b": row["b"]}),
                weave.legacy.weave.ops.dict_(**{"a": row["val_a"], "b": row["val_b"]}),
            )
        )
    ).to_pylist_raw() == [5, None, 10]


def test_cond_vector_arr_value():
    conds = weave.save(
        ops_arrow.to_arrow(
            [
                {"a": True, "b": False, "val_a": [1, 2], "val_b": [3, 4]},
                {"a": False, "b": False, "val_a": [5, 6], "val_b": [7, 8]},
                {"a": False, "b": True, "val_a": [9, 10], "val_b": [11, 12]},
            ]
        ),
        "conds",
    )
    assert weave.use(
        conds.map(
            lambda row: weave.legacy.weave.ops.cond(
                weave.legacy.weave.ops.dict_(**{"a": row["a"], "b": row["b"]}),
                weave.legacy.weave.ops.dict_(**{"a": row["val_a"], "b": row["val_b"]}),
            )
        )
    ).to_pylist_raw() == [[1, 2], None, [11, 12]]


def test_cond_vector_mixed():
    conds = weave.save(
        ops_arrow.to_arrow(
            [
                {"a": True, "b": False, "val_a": 1, "val_b": 4},
                {"a": False, "b": False, "val_a": 2, "val_b": 5},
                {"a": False, "b": True, "val_a": 3, "val_b": 6},
            ]
        ),
        "conds",
    )
    assert weave.use(
        conds.map(
            lambda row: weave.legacy.weave.ops.cond(
                weave.legacy.weave.ops.dict_(**{"a": row["a"], "b": row["b"]}),
                weave.legacy.weave.ops.dict_(**{"a": row["val_a"], "b": 99}),
            )
        )
    ).to_pylist_raw() == [1, None, 99]


def test_cond_vector_mixed_arr_value():
    conds = weave.save(
        ops_arrow.to_arrow(
            [
                {"a": True, "b": False, "val_a": [1, 2], "val_b": [3, 4]},
                {"a": False, "b": False, "val_a": [5, 6], "val_b": [7, 8]},
                {"a": False, "b": True, "val_a": [9, 10], "val_b": [11, 12]},
            ]
        ),
        "conds",
    )
    assert weave.use(
        conds.map(
            lambda row: weave.legacy.weave.ops.cond(
                weave.legacy.weave.ops.dict_(**{"a": row["a"], "b": row["b"]}),
                weave.legacy.weave.ops.dict_(**{"a": row["val_a"], "b": [99, 100]}),
            )
        )
    ).to_pylist_raw() == [[1, 2], None, [99, 100]]



import typing

import openai
from openai._types import NotGiven
from openai.types.chat import (
    ChatCompletionMessageParam,
)
from pydantic import Field

import weave
from weave.flow.chat_util import OpenAIStream
from weave.flow.console import LogEvents
from weave.flow.obj import Object
from weave.flow.tools import chat_call_tool_params, perform_tool_calls


class AgentState(Object):
    # TODO: want openai types here.
    history: list[typing.Any] = Field(default_factory=list)


class Agent(Object):
    model_name: str = "gpt-3.5-turbo"
    temperature: float = 0.7
    system_message: str
    tools: list[typing.Any] = Field(default_factory=list)

    @weave.op()
    def step(self, state: AgentState) -> AgentState:
        """Run a step of the agent.

        Args:
            state: The current state of the environment.
            action: The action to take.

        Returns:
            The new state of the environment.
        """
        LogEvents.step_start("agent", "green")

        messages: list[ChatCompletionMessageParam] = [
            {"role": "system", "content": self.system_message},
        ]
        messages += state.history

        # make type checkers happy by passing NotGiven instead of None
        tools = NotGiven()
        if self.tools:
            tools = chat_call_tool_params(self.tools)

        LogEvents.chat_response_start()
        stream = openai.chat.completions.create(
            model=self.model_name,
            temperature=self.temperature,
            messages=messages,
            tools=tools,
            stream=True,
        )
        wrapped_stream = OpenAIStream(stream)
        # with console.status("Streaming LLM response..."):
        for chunk in wrapped_stream:
            if chunk.choices[0].delta.content:
                LogEvents.chat_message_content_delta(chunk.choices[0].delta.content)

        response = wrapped_stream.final_response()
        response_message = response.choices[0].message
        if response_message.content:
            LogEvents.chat_response_complete(response_message.content)

        new_messages = []
        # we always store the dict representations of messages in agent state
        # instead of mixing in some pydantic objects.
        new_messages.append(response_message.model_dump(exclude_none=True))
        if response_message.tool_calls:
            new_messages.extend(
                perform_tool_calls(self.tools, response_message.tool_calls)
            )

        # TODO: doesn't really belong here.
        # find distance from end to last system message in state.history
        # if distance > 10, append system message
        for distance in range(len(state.history)):
            if state.history[-(distance + 1)]["role"] == "system":
                break
        if distance > 10:
            new_messages.append({"role": "system", "content": self.system_message})

        return AgentState(history=state.history + new_messages)



import asyncio
import multiprocessing
from typing import Any, AsyncIterator, Awaitable, Callable, Iterable, Tuple, TypeVar

T = TypeVar("T")
U = TypeVar("U")


async def async_foreach(
    sequence: Iterable[T],
    func: Callable[[T], Awaitable[U]],
    max_concurrent_tasks: int,
) -> AsyncIterator[Tuple[T, U]]:
    semaphore = asyncio.Semaphore(max_concurrent_tasks)

    async def process_item(item: T) -> Tuple[T, U]:
        async with semaphore:
            result = await func(item)
            return item, result

    tasks = [asyncio.create_task(process_item(item)) for item in sequence]

    for task in asyncio.as_completed(tasks):
        item, result = await task
        yield item, result


def _subproc(
    queue: multiprocessing.Queue, func: Callable, *args: Any, **kwargs: Any
) -> None:
    result = func(*args, **kwargs)
    queue.put(result)


def _run_in_process(
    func: Callable, args: Tuple = (), kwargs: dict = {}
) -> Tuple[multiprocessing.Process, multiprocessing.Queue]:
    """Run a function in a separate process and return the process object and a multiprocessing.Queue for the result."""
    queue: multiprocessing.Queue = multiprocessing.Queue()
    process: multiprocessing.Process = multiprocessing.Process(
        target=_subproc, args=(queue, func) + args, kwargs=kwargs
    )
    process.start()
    return process, queue


async def run_in_process_with_timeout(
    timeout: float, func: Callable, *args: Any, **kwargs: Any
) -> Any:
    """Run a function in a separate process with a timeout. Terminate the process if it exceeds the timeout."""
    # Note, on osx, multiprocessing uses spawn by default. With span, subprocesses will re-import main,
    # and incur the cost of bootstrapping python and all imports, which for weave is currently about 1.5s.
    # Scripts that use this can call 'multiprocessing.set_start_method("fork")' which is much faster (0.05s)
    # depending on their use case (there are some issues with fork on osx).
    loop = asyncio.get_running_loop()
    process, queue = _run_in_process(func, args, kwargs)

    try:
        # Wait for the process to complete or timeout
        await asyncio.wait_for(loop.run_in_executor(None, process.join), timeout)
    except asyncio.TimeoutError:
        print("Function execution exceeded the timeout. Terminating process.")
        process.terminate()
        process.join()  # Ensure process resources are cleaned up
        raise

    if process.exitcode == 0:
        return queue.get()  # Retrieve result from the queue
    else:
        raise ValueError(
            "Unhandled exception in subprocess. Exitcode: " + str(process.exitcode)
        )
